from bundlewrap.utils import Fault

from pytest import raises


def test_basic_resolve():
    def callback():
        return 4  # Chosen by fair dice roll. Guaranteed to be random.

    f = Fault('id', callback)
    assert f.value == 4


def test_add_fault():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'bar'

    a = Fault('id foo', callback_a)
    b = Fault('id bar', callback_b)
    c = a + b
    assert c.value == 'foobar'


def test_add_fault_nonstring():
    def callback_a():
        return 4
    def callback_b():
        return 8

    a = Fault('id foo', callback_a)
    b = Fault('id bar', callback_b)
    c = a + b
    assert c.value == 12


def test_add_plain_nonstring():
    def callback():
        return 4

    a = Fault('id foo', callback)
    b = a + 8
    assert b.value == 12


def test_add_plain():
    def callback_a():
        return 'foo'

    a = Fault('id foo', callback_a)
    c = a + 'bar'
    assert c.value == 'foobar'


def test_order():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'bar'
    def callback_c():
        return '0first'

    a = Fault('id foo', callback_a)
    b = Fault('id bar', callback_b)
    c = Fault('id 0first', callback_c)

    lst = sorted([a, b, c])

    assert lst[0].value == '0first'
    assert lst[1].value == 'bar'
    assert lst[2].value == 'foo'


def test_b64encode():
    def callback():
        return 'foo'

    a = Fault('id foo', callback).b64encode()
    assert a.value == 'Zm9v'


def test_format_into():
    def callback():
        return 'foo'

    a = Fault('id foo', callback).format_into('This is my secret: "{}"')
    assert a.value == 'This is my secret: "foo"'


# XXX Other methods missing. This basically tests if
# _make_method_callback() is working.
def test_generic_method_lower():
    def callback():
        return 'FOO'

    a = Fault('id FOO', callback)
    assert a.lower().value == 'foo'


def test_equal_no_operators():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo, but here you see the problem'

    a = Fault('id foo', callback_a)
    b = Fault('id foo', callback_b)
    assert id(a) != id(b)
    assert a == b


def test_not_equal_no_operators():
    def callback_a():
        return 'this interface is not fool proof'
    def callback_b():
        return 'this interface is not fool proof'

    a = Fault('id foo', callback_a)
    b = Fault('id bar', callback_b)
    assert id(a) != id(b)
    assert a != b


def test_equal_lower():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).lower()
    b = Fault('id foo', callback_b).lower()
    assert id(a) != id(b)
    assert a == b


def test_not_equal_lower():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).lower()
    b = Fault('id bar', callback_b).lower()
    assert id(a) != id(b)
    assert a != b


def test_equal_b64encode():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).b64encode()
    b = Fault('id foo', callback_b).b64encode()
    assert id(a) != id(b)
    assert a == b


def test_not_equal_b64encode():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).b64encode()
    b = Fault('id bar', callback_b).b64encode()
    assert id(a) != id(b)
    assert a != b


def test_equal_format_into():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).format_into('bar {}')
    b = Fault('id foo', callback_b).format_into('bar {}')
    assert id(a) != id(b)
    assert a == b


def test_not_equal_format_into():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).format_into('bar {}')
    b = Fault('id foo', callback_b).format_into('baz {}')
    assert id(a) != id(b)
    assert a != b


def test_nested_equal():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).lower().b64encode()
    b = Fault('id foo', callback_b).lower().b64encode()
    assert id(a) != id(b)
    assert a == b


def test_nested_not_equal_because_of_id():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).lower().b64encode()
    b = Fault('id bar', callback_b).lower().b64encode()
    assert id(a) != id(b)
    assert a != b


def test_nested_not_equal_because_of_operators():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'foo'

    a = Fault('id foo', callback_a).lower().b64encode()
    b = Fault('id foo', callback_b).lower()
    assert id(a) != id(b)
    assert a != b


def test_can_be_used_in_set():
    def callback_a():
        return 'foo'
    def callback_b():
        return 'bar'

    a = Fault('id foo', callback_a)
    b = Fault('id bar', callback_b)
    s = {a, a, b}
    assert len(s) == 2
    assert 'foo' in [i.value for i in s]
    assert 'bar' in [i.value for i in s]


def test_kwargs_add_to_idlist():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz='bam', frob='glob')
    b = Fault('id foo', callback, different='kwargs')
    assert a != b
    assert hash(a) != hash(b)


def test_unhashable_dict_kwargs_add_to_idlist():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz={1: {2: {3: 4}}})
    b = Fault('id foo', callback, foo='bar', baz={1: {3: {3: 4}}})
    assert a != b
    assert hash(a) != hash(b)


def test_unhashable_list_kwargs_add_to_idlist():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz=[1, 2, [3, 4]])
    b = Fault('id foo', callback, foo='bar', baz=[1, [3, 4], 2])
    assert a != b
    assert hash(a) != hash(b)


def test_unhashable_set_kwargs_add_to_idlist():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz={1, 2, 3})
    b = Fault('id foo', callback, foo='bar', baz={1, 2, 4})
    assert a != b
    assert hash(a) != hash(b)


def test_unhashable_dict_kwargs_add_to_idlist_equal():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz={1: {2: {3: 4, 5: 6}}})
    b = Fault('id foo', callback, foo='bar', baz={1: {2: {5: 6, 3: 4}}})
    assert a == b
    assert hash(a) == hash(b)


def test_unhashable_list_kwargs_add_to_idlist_equal():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz=[1, 2, 3])
    b = Fault('id foo', callback, foo='bar', baz=[1, 2, 3])
    assert id(a) != id(b)
    assert a == b


def test_unhashable_set_kwargs_add_to_idlist_equal():
    def callback():
        return 'foo'

    a = Fault('id foo', callback, foo='bar', baz={1, 2, 3})
    b = Fault('id foo', callback, foo='bar', baz={1, 3, 2})
    assert a == b
    assert hash(a) == hash(b)


def test_eq_and_hash_do_not_resolve_fault():
    def callback():
        raise Exception('Fault resolved, this should not happen')

    a = Fault('id foo', callback)
    b = Fault('id foo', callback)
    assert a == b

    s = {a, b}


def test_kwargs_changed_after_creation():
    def callback():
        return 'foo'

    data = {
        'foo': 0,
    }
    a = Fault('id foo', callback, data=data)

    data['foo'] = 1
    b = Fault('id foo', callback, data=data)

    # Even though both Faults reference the same dict, hashes are built
    # on Fault creation based on the actual values in mutable
    # parameters.
    assert a != b
    assert hash(a) != hash(b)


def test_kwargs_not_changed_after_creation():
    def callback():
        return 'foo'

    data = {
        'foo': 0,
    }
    a = Fault('id foo', callback, data=data)
    b = Fault('id foo', callback, data=data)

    assert a == b
    assert hash(a) == hash(b)


def test_hash_does_not_change():
    def callback():
        return 'foo'

    data = {
        'foo': 0,
    }
    a = Fault('id foo', callback, data=data)
    hash1 = hash(a)

    data['foo'] = 1
    hash2 = hash(a)

    assert hash1 == hash2


def test_sort():
    def one():
        return 1

    def three():
        return 3

    f1 = Fault("1", one)
    f3 = Fault("3", three)

    assert sorted([2, f3, f1]) == [f1, 2, f3]


def test_sort_typeerror():
    def one():
        return 1

    def three():
        return 3

    f1 = Fault("1", one)
    f3 = Fault("3", three)

    with raises(TypeError):
        sorted(["2", f3, f1])


def test_sort_typeerror_from_fault():
    def one():
        return 1

    def three():
        return "3"

    f1 = Fault("1", one)
    f3 = Fault("3", three)

    with raises(TypeError):
        sorted([2, f3, f1])



from os import environ
from sys import exit

from ..utils import names
from ..utils.cmdline import get_target_nodes
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import bold, mark_for_translation as _, red
from ..utils.ui import io, page_lines
from ..group import GROUP_ATTR_DEFAULTS


NODE_ATTRS = sorted(list(GROUP_ATTR_DEFAULTS) + ['bundles', 'file_path', 'groups', 'hostname'])
NODE_ATTRS_LISTS = ('bundles', 'groups')


def _attribute_table(
    entities,
    entity_label,
    selected_attrs,
    available_attrs,
    available_attrs_lists,
    inline,
):
    rows = [[entity_label], ROW_SEPARATOR]
    selected_attrs = [attr.strip() for attr in selected_attrs]

    if selected_attrs == ['all']:
        selected_attrs = available_attrs
    elif 'all' in selected_attrs:
        io.stderr(_(
            "{x} invalid attribute list requested ('all' and extraneous): {attr}"
        ).format(x=red("!!!"), attr=", ".join(sorted(selected_attrs))))
        exit(1)

    for attr in selected_attrs:
        if attr not in available_attrs:
            io.stderr(_("{x} unknown attribute: {attr}").format(x=red("!!!"), attr=attr))
            exit(1)
        rows[0].append(bold(attr))

    has_list_attrs = False
    for entity in sorted(entities):
        attr_values = [[entity.name]]
        for attr in selected_attrs:
            if attr in available_attrs_lists:
                if inline:
                    attr_values.append([",".join(sorted(names(getattr(entity, attr))))])
                else:
                    has_list_attrs = True
                    attr_values.append(sorted(names(getattr(entity, attr))))
            else:
                attr_values.append([str(getattr(entity, attr))])
        number_of_lines = max([len(value) for value in attr_values])
        if environ.get("BW_TABLE_STYLE") == 'grep':
            # repeat entity name for each line
            attr_values[0] = attr_values[0] * number_of_lines
        for line in range(number_of_lines):
            row = []
            for attr_index in range(len(selected_attrs) + 1):
                try:
                    row.append(attr_values[attr_index][line])
                except IndexError:
                    row.append("")
            rows.append(row)
        if has_list_attrs:
            rows.append(ROW_SEPARATOR)
    if environ.get("BW_TABLE_STYLE") == 'grep':
        rows = rows[2:]
    page_lines(render_table(
        rows[:-1] if has_list_attrs else rows,  # remove trailing ROW_SEPARATOR
    ))


def bw_nodes(repo, args):
    if args['targets']:
        nodes = get_target_nodes(repo, args['targets'])
    else:
        nodes = repo.nodes
    if not args['attrs']:
        for node in sorted(nodes):
            io.stdout(node.name)
    else:
        _attribute_table(
            nodes,
            bold(_("node")),
            args['attrs'],
            NODE_ATTRS,
            NODE_ATTRS_LISTS,
            args['inline'],
        )



from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_not_a_repo_test(tmpdir):
    assert run("bw nodes", path=str(tmpdir))[2] == 1


def test_subdir_invocation(tmpdir):
    make_repo(tmpdir, nodes={"node1": {}})
    stdout, stderr, rcode = run("bw nodes", path=join(str(tmpdir), "bundles"))
    assert stdout == b"node1\n"
    assert stderr == b""
    assert rcode == 0



from os.path import exists, join

from bundlewrap.utils.testing import host_os, make_repo, run


def test_deploy_from_url(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'git_deploy': {
                        join(str(tmpdir), "git_deployed_bw"): {
                            'repo': "https://github.com/bundlewrap/bundlewrap.git",
                            'rev': "master",
                        },
                    },
                    'directories': {
                        join(str(tmpdir), "git_deployed_bw"): {},
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )

    assert not exists(join(str(tmpdir), "git_deployed_bw", "LICENSE"))
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert exists(join(str(tmpdir), "git_deployed_bw", "LICENSE"))
    assert not exists(join(str(tmpdir), "git_deployed_bw", ".git"))


def test_cannot_deploy_into_purged(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'git_deploy': {
                        join(str(tmpdir), "git_deployed_bw"): {
                            'repo': "https://github.com/bundlewrap/bundlewrap.git",
                            'rev': "master",
                        },
                    },
                    'directories': {
                        join(str(tmpdir), "git_deployed_bw"): {
                            'purge': True,
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )

    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 1
    assert b"cannot git_deploy into purged directory" in stderr



from json import loads
from os import environ

from bundlewrap.utils.testing import host_os, make_repo, run


if environ.get('TRAVIS') == "true":
    def test_create(tmpdir):
        make_repo(
            tmpdir,
            bundles={
                "test": {
                    'items': {
                        'postgres_dbs': {
                            "bw-test1": {
                                'owner': "bw-test1",
                            },
                        },
                        'postgres_roles': {
                            "bw-test1": {
                                'superuser': True,
                                'password': 'potato',
                            },
                        },
                    },
                },
            },
            nodes={
                "localhost": {
                    'bundles': ["test"],
                    'os': host_os(),
                },
            },
        )

        stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
        assert rcode == 0

        stdout, stderr, rcode = run("bw items --state localhost postgres_db:bw-test1", path=str(tmpdir))
        assert rcode == 0
        assert loads(stdout.decode()) == {'owner': "bw-test1"}

        stdout, stderr, rcode = run("bw items --state localhost postgres_role:bw-test1", path=str(tmpdir))
        assert rcode == 0
        assert loads(stdout.decode()) == {
            'can_login': True,
            'password_hash': "md5ecba3aec62c5aabf6480de6352182004",
            'superuser': True,
        }

        stdout, stderr, rcode = run("dropdb bw-test1", path=str(tmpdir))
        assert rcode == 0
        stdout, stderr, rcode = run("dropuser bw-test1", path=str(tmpdir))
        assert rcode == 0



from datetime import datetime
from itertools import zip_longest
from sys import exit

from ..concurrency import WorkerPool
from ..exceptions import SkipNode
from ..utils import SkipList
from ..utils.cmdline import get_target_nodes
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import (
    blue,
    bold,
    error_summary,
    format_duration,
    green,
    mark_for_translation as _,
    red,
    yellow,
)
from ..utils.ui import io


def run_on_node(node, command, skip_list):
    if node.dummy:
        io.stdout(_("{x} {node}  is a dummy node").format(node=bold(node.name), x=yellow("»")))
        return None

    if node.name in skip_list:
        io.stdout(_("{x} {node}  skipped by --resume-file").format(node=bold(node.name), x=yellow("»")))
        return None

    try:
        node.repo.hooks.node_run_start(
            node.repo,
            node,
            command,
        )
    except SkipNode as exc:
        io.stdout(_("{x} {node}  skipped by hook ({reason})").format(
            node=bold(node.name),
            reason=str(exc) or _("no reason given"),
            x=yellow("»"),
        ))
        return None

    with io.job(_("{}  running command...").format(bold(node.name))):
        result = node.run(
            command,
            may_fail=True,
            log_output=True,
        )

    node.repo.hooks.node_run_end(
        node.repo,
        node,
        command,
        duration=result.duration,
        return_code=result.return_code,
        stdout=result.stdout,
        stderr=result.stderr,
    )
    return result


def stats_summary(results, include_stdout, include_stderr):
    rows = [[
        bold(_("node")),
        bold(_("return code")),
        bold(_("time")),
    ], ROW_SEPARATOR]
    if include_stdout:
        rows[0].append(bold(_("stdout")))
    if include_stderr:
        rows[0].append(bold(_("stderr")))

    for node_name, result in sorted(results.items()):
        row = [node_name]
        if result is None:
            # node has been skipped
            continue
        elif result.return_code == 0:
            row.append(green(str(result.return_code)))
        else:
            row.append(red(str(result.return_code)))
        row.append(format_duration(result.duration, msec=True))
        rows.append(row)
        if include_stdout or include_stderr:
            stdout = result.stdout.decode('utf-8', errors='replace').strip().split("\n")
            stderr = result.stderr.decode('utf-8', errors='replace').strip().split("\n")
            if include_stdout:
                row.append(stdout[0])
            if include_stderr:
                row.append(stderr[0])
            for stdout_line, stderr_line in list(zip_longest(stdout, stderr, fillvalue=""))[1:]:
                continuation_row = ["", "", ""]
                if include_stdout:
                    continuation_row.append(stdout_line)
                if include_stderr:
                    continuation_row.append(stderr_line)
                rows.append(continuation_row)
            rows.append(ROW_SEPARATOR)

    if include_stdout or include_stderr:
        # remove last ROW_SEPARATOR
        rows = rows[:-1]
    if len(rows) > 2:  # table might be empty if all nodes have been skipped
        for line in render_table(rows, alignments={1: 'right', 2: 'right'}):
            io.stdout("{x} {line}".format(x=blue("i"), line=line))


def bw_run(repo, args):
    errors = []
    target_nodes = get_target_nodes(repo, args['targets'])
    pending_nodes = target_nodes.copy()
    io.progress_set_total(len(pending_nodes))

    repo.hooks.run_start(
        repo,
        args['targets'],
        target_nodes,
        args['command'],
    )
    start_time = datetime.now()
    results = {}
    skip_list = SkipList(args['resume_file'])

    def tasks_available():
        return bool(pending_nodes)

    def next_task():
        node = pending_nodes.pop()
        return {
            'target': run_on_node,
            'task_id': node.name,
            'args': (
                node,
                args['command'],
                skip_list,
            ),
        }

    def handle_result(task_id, return_value, duration):
        io.progress_advance()
        results[task_id] = return_value
        if return_value is None or return_value.return_code == 0:
            skip_list.add(task_id)

    def handle_exception(task_id, exception, traceback):
        io.progress_advance()
        msg = "{}  {}".format(bold(task_id), exception)
        io.stderr(traceback)
        io.stderr(repr(exception))
        io.stderr("{} {}".format(red("!"), msg))
        errors.append(msg)

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result=handle_result,
        handle_exception=handle_exception,
        cleanup=skip_list.dump,
        pool_id="run",
        workers=args['node_workers'],
    )
    worker_pool.run()

    if args['summary']:
        stats_summary(results, args['stdout_table'], args['stderr_table'])
    error_summary(errors)

    repo.hooks.run_end(
        repo,
        args['targets'],
        target_nodes,
        args['command'],
        duration=datetime.now() - start_time,
    )

    exit(1 if errors else 0)



from copy import copy
from hashlib import sha1
from json import dumps, JSONEncoder

from .exceptions import RepositoryError
from .utils import Fault
from .utils.dicts import ATOMIC_TYPES, map_dict_keys, merge_dict, value_at_key_path
from .utils.text import force_text, mark_for_translation as _


METADATA_TYPES = (  # only meant for natively atomic types
    bool,
    bytes,
    Fault,
    int,
    str,
    type(None),
)


class DoNotRunAgain(Exception):
    """
    Raised from metadata reactors to indicate they can be disregarded.
    """
    pass


def deepcopy_metadata(obj):
    """
    Our own version of deepcopy.copy that doesn't pickle.
    """
    if isinstance(obj, METADATA_TYPES):
        return obj
    elif isinstance(obj, dict):
        if isinstance(obj, ATOMIC_TYPES[dict]):
            new_obj = atomic({})
        else:
            new_obj = {}
        for key, value in obj.items():
            new_key = copy(key)
            new_obj[new_key] = deepcopy_metadata(value)
    elif isinstance(obj, (list, tuple)):
        if isinstance(obj, (ATOMIC_TYPES[list], ATOMIC_TYPES[tuple])):
            new_obj = atomic([])
        else:
            new_obj = []
        for member in obj:
            new_obj.append(deepcopy_metadata(member))
    elif isinstance(obj, set):
        if isinstance(obj, ATOMIC_TYPES[set]):
            new_obj = atomic(set())
        else:
            new_obj = set()
        for member in obj:
            new_obj.add(deepcopy_metadata(member))
    else:
        assert False  # there should be no other types
    return new_obj


def validate_metadata(metadata, _top_level=True):
    if _top_level and not isinstance(metadata, dict):
        raise TypeError(_("metadata must be a dict"))
    if isinstance(metadata, dict):
        for key, value in metadata.items():
            if not isinstance(key, str):
                raise TypeError(_("metadata keys must be str: {value} is {type}").format(
                    type=type(key),
                    value=repr(key),
                ))
            validate_metadata(value, _top_level=False)
    elif isinstance(metadata, (tuple, list, set)):
        for value in metadata:
            validate_metadata(value, _top_level=False)
    elif not isinstance(metadata, METADATA_TYPES):
        raise TypeError(_("illegal metadata value type: {value} is {type}").format(
            type=type(metadata),
            value=repr(metadata),
        ))


def atomic(obj):
    """
    Wraps a compatible object in a custom class to prevent it from being
    merged with another object of the same type during metadata
    compilation.
    """
    try:
        cls = ATOMIC_TYPES[type(obj)]
    except KeyError:
        raise ValueError("atomic() can only be applied to dicts, lists, sets, or tuples "
                         "(not: {})".format(repr(obj)))
    else:
        return cls(obj)


def check_for_metadata_conflicts(node):
    check_for_metadata_conflicts_between_groups(node)
    check_for_metadata_conflicts_between_defaults_and_reactors(node)


def check_for_metadata_conflicts_between_defaults_and_reactors(node):
    """
    Finds conflicting metadata keys in bundle defaults and reactors.

    Dicts can be merged with dicts, sets can be merged with sets, but
    any other combination is a conflict.
    """
    TYPE_DICT = 1
    TYPE_SET = 2
    TYPE_OTHER = 3

    def paths_with_values_and_types(d):
        for path in map_dict_keys(d):
            value = value_at_key_path(d, path)
            if isinstance(value, dict):
                yield path, value, TYPE_DICT
            elif isinstance(value, set):
                yield path, value, TYPE_SET
            else:
                yield path, value, TYPE_OTHER

    for prefix in ("metadata_defaults:", "metadata_reactor:"):
        paths = {}
        node.metadata.get(tuple())  # ensure full metadata is present
        for partition in node.metadata.stack._partitions:
            for identifier, layer in partition.items():
                if identifier.startswith(prefix):
                    for path, value, current_type in paths_with_values_and_types(layer):
                        try:
                            prev_type, prev_identifier, prev_value = paths[path]
                        except KeyError:
                            paths[path] = current_type, identifier, value
                        else:
                            if (
                                prev_type == TYPE_DICT
                                and current_type == TYPE_DICT
                            ):
                                pass
                            elif (
                                prev_type == TYPE_SET
                                and current_type == TYPE_SET
                            ):
                                pass
                            elif value != prev_value:
                                raise ValueError(_(
                                    "{node}: {a} and {b} are clashing over this key path: {path} "
                                    "({val_a} vs. {val_b})"
                                ).format(
                                    a=identifier,
                                    b=prev_identifier,
                                    node=node.name,
                                    path="/".join(path),
                                    val_a=repr(value),
                                    val_b=repr(prev_value),
                                ))


def check_for_metadata_conflicts_between_groups(node):
    """
    Finds metadata keys defined by two groups that are not part of a
    shared subgroup hierarchy.
    """
    # First, we build a list of subgroup chains.
    #
    # A chain is simply a list of groups starting with a parent group
    # that has no parent groups itself and then descends depth-first
    # into its subgroups until a subgroup is reached that the node is
    # not a member of.
    # Every possible path on every subgroup tree is a separate chain.
    #
    #      group4
    #     /     \
    #  group2  group3
    #     \     /
    #     group1
    #
    # This example has two chains, even though both start and end at the
    # some groups:
    #
    #     group1 -> group2 -> group4
    #     group1 -> group3 -> group4
    #

    # find all groups whose subgroups this node is *not* a member of
    lowest_subgroups = set()
    for group in node.groups:
        in_subgroup = False
        for subgroup in group.subgroups:
            if subgroup in node.groups:
                in_subgroup = True
                break
        if not in_subgroup:
            lowest_subgroups.add(group)

    chains = []
    incomplete_chains = [[group] for group in lowest_subgroups]

    while incomplete_chains:
        for chain in incomplete_chains[:]:
            highest_group = chain[-1]
            if list(highest_group.parent_groups):
                chain_so_far = chain[:]
                # continue this chain with the first parent group
                chain.append(list(highest_group.parent_groups)[0])
                # further parent groups form new chains
                for further_parents in list(highest_group.parent_groups)[1:]:
                    new_chain = chain_so_far[:]
                    new_chain.append(further_parents)
                    incomplete_chains.append(new_chain)
            else:
                # chain has ended
                chains.append(chain)
                incomplete_chains.remove(chain)

    # chains now look like this (parents right of children):
    # [
    #     [group1],
    #     [group2, group3, group5],
    #     [group2, group4, group5],
    #     [group2, group4, group6, group7],
    # ]

    # let's merge metadata for each chain
    chain_metadata = []
    for chain in chains:
        metadata = {}
        for group in chain:
            metadata = merge_dict(metadata, group._attributes.get('metadata', {}))
        chain_metadata.append(metadata)

    # create a "key path map" for each chain's metadata
    chain_metadata_keys = [list(map_dict_keys(metadata)) for metadata in chain_metadata]

    # compare all metadata keys with other chains and find matches
    for index1, keymap1 in enumerate(chain_metadata_keys):
        for keypath in keymap1:
            for index2, keymap2 in enumerate(chain_metadata_keys):
                if index1 == index2:
                    # same keymap, don't compare
                    continue
                else:
                    if keypath in keymap2:
                        if (
                            type(value_at_key_path(chain_metadata[index1], keypath)) ==
                            type(value_at_key_path(chain_metadata[index2], keypath)) and
                            type(value_at_key_path(chain_metadata[index2], keypath)) in
                            (set, dict)
                        ):
                            continue
                        # We now know that there is a conflict between the first
                        # and second chain we're looking at right now.
                        # That is however not a problem if the conflict is caused
                        # by a group that is present in both chains.
                        # So all that's left is to figure out which two single groups
                        # within those chains are at fault so we can report them
                        # to the user if necessary.
                        find_groups_causing_metadata_conflict(
                            node.name,
                            chains[index1],
                            chains[index2],
                            keypath,
                        )


def find_groups_causing_metadata_conflict(node_name, chain1, chain2, keypath):
    """
    Given two chains (lists of groups), find one group in each chain
    that has conflicting metadata with the other for the given key path.
    """
    chain1_metadata = [
        list(map_dict_keys(group._attributes.get('metadata', {}))) for group in chain1
    ]
    chain2_metadata = [
        list(map_dict_keys(group._attributes.get('metadata', {}))) for group in chain2
    ]

    bad_keypath = None

    for index1, keymap1 in enumerate(chain1_metadata):
        for index2, keymap2 in enumerate(chain2_metadata):
            if chain1[index1] == chain2[index2]:
                # same group, ignore
                continue
            if (
                keypath in keymap1 and
                keypath in keymap2 and
                chain1[index1] not in chain2[index2].subgroups and
                chain2[index2] not in chain1[index1].subgroups
            ):
                bad_keypath = keypath
                bad_group1 = chain1[index1]
                bad_group2 = chain2[index2]

    if bad_keypath is not None:
        raise RepositoryError(_(
            "Conflicting metadata keys between groups '{group1}' and '{group2}' on node '{node}':\n\n"
            "    metadata['{keypath}']\n\n"
            "You must either connect both groups through subgroups or have them not define "
            "conflicting metadata keys. Otherwise there is no way for BundleWrap to determine "
            "which group's metadata should win when they are merged."
        ).format(
            keypath="']['".join(bad_keypath),
            group1=bad_group1.name,
            group2=bad_group2.name,
            node=node_name,
        ))


class MetadataJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Fault):
            return obj.value
        if isinstance(obj, set):
            return sorted(obj)
        if isinstance(obj, bytes):
            return force_text(obj)
        else:
            raise ValueError(_("illegal metadata value type: {value} is {type}").format(
                type=type(obj),
                value=repr(obj),
            ))


def metadata_to_json(metadata, sort_keys=True):
    if not isinstance(metadata, dict):  # might be NodeMetadataProxy
        metadata = dict(metadata)
    return dumps(
        metadata,
        cls=MetadataJSONEncoder,
        indent=4,
        sort_keys=sort_keys,
    )


def hash_metadata(sdict):
    """
    Returns a canonical SHA1 hash to describe this dict.
    """
    return sha1(metadata_to_json(sdict).encode('utf-8')).hexdigest()



from . import names
from .text import mark_for_translation as _, red


def explain_item_dependency_loop(exc, node_name):
    """
    Given an ItemDependencyLoop exception and a node name, generates
    output lines to help users debug the issue.
    """
    items = remove_items_not_contributing_to_loop(exc.items)

    yield _(
        "{x} There was a dependency problem on node '{node}'. Look at the debug.svg generated "
        "by the following command and try to find a loop:\n\n\n"
        "printf '{cmd}' | dot -Tsvg -odebug.svg\n\n\n"
    ).format(
        x=red("!"),
        node=node_name,
        cmd="\\n".join(graph_for_items(node_name, items)),
    )

    yield _(
        "{x} Additionally, here is a list of all items involved "
        "and their remaining dependencies:\n"
    ).format(x=red("!"))
    for item in items:
        yield "{}\t{}".format(item.id, ",".join([item.id for item in sorted(item._deps)]))
    yield "\n\n\n"


def graph_for_items(
    title,
    items,
    cluster=True,
    concurrency=True,
    regular=True,
    reverse=True,
    auto=True,
):
    items = sorted(items)

    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("graph [color=\"#303030\"; "
                  "fontname=Helvetica; "
                  "penwidth=2; "
                  "shape=box; "
                  "style=\"rounded,dashed\"]")
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontcolor=white; "
                 "fontname=Helvetica; "
                 "shape=box; "
                 "style=\"rounded,filled\"]")
    yield "edge [arrowhead=vee]"

    item_ids = []
    for item in items:
        item_ids.append(item.id)

    if cluster:
        # Define which items belong to which bundle
        bundle_number = 0
        bundles_seen = set()
        for item in items:
            if item.bundle is None or item.bundle.name in bundles_seen:
                continue
            yield "subgraph cluster_{}".format(bundle_number)
            bundle_number += 1
            yield "{"
            yield "label = \"{}\"".format(item.bundle.name)
            if "bundle:{}".format(item.bundle.name) in item_ids:
                yield "\"bundle:{}\"".format(item.bundle.name)
            for bitem in item.bundle.items:
                if bitem.id in item_ids:
                    yield "\"{}\"".format(bitem.id)
            yield "}"
            bundles_seen.add(item.bundle.name)

    # Define dependencies between items
    for item in items:
        if regular:
            for dep in item.needs:
                if dep in item_ids:
                    yield "\"{}\" -> \"{}\" [color=\"#C24948\",penwidth=2]".format(item.id, dep)

        if auto:
            for dep in sorted(item._deps):
                if dep not in items:
                    continue
                if dep.id in getattr(item, '_concurrency_deps', []):
                    if concurrency:
                        yield "\"{}\" -> \"{}\" [color=\"#714D99\",penwidth=2]".format(item.id, dep)
                elif dep in item._reverse_deps:
                    if reverse:
                        yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(item.id, dep)
                elif dep.id not in item.needs:
                    if dep in items:
                        yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(item.id, dep)

    # Global graph title
    yield "fontsize = 28"
    yield "label = \"{}\"".format(title)
    yield "labelloc = \"t\""
    yield "}"


def plot_group(groups, nodes, show_nodes):
    groups = sorted(groups)
    nodes = sorted(nodes)

    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontname=Helvetica]")
    yield "edge [arrowhead=vee]"

    for group in groups:
        yield "\"{}\" [fontcolor=white,style=filled];".format(group.name)

    for node in nodes:
        yield "\"{}\" [fontcolor=\"#303030\",shape=box,style=rounded];".format(node.name)

    for group in groups:
        for subgroup in sorted(group._attributes.get('subgroups', set())):
            yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(group.name, subgroup)
        for subgroup in sorted(group._subgroup_names_from_patterns):
            yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(group.name, subgroup)

    if show_nodes:
        for group in groups:
            for node in nodes:
                if group in set(node._attributes.get('groups', set())):
                    yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                        node.name, group.name)
                elif node in group._nodes_from_members:
                    yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                        group.name, node.name)
                else:
                    for pattern in sorted(group._member_patterns):
                        if pattern.search(node.name) is not None:
                            yield "\"{}\" -> \"{}\" [color=\"#714D99\",penwidth=2]".format(
                                group.name, node.name)
                            break
    yield "}"


def plot_node_groups(node):
    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontname=Helvetica]")
    yield "edge [arrowhead=vee]"

    for group in sorted(node.groups):
        yield "\"{}\" [fontcolor=white,style=filled];".format(group.name)

    yield "\"{}\" [fontcolor=\"#303030\",shape=box,style=rounded];".format(node.name)

    for group in sorted(node.groups):
        for subgroup in sorted(group._attributes.get('subgroups', set())):
            if subgroup in names(node.groups):
                yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(
                    group.name, subgroup)
        for pattern in sorted(group._immediate_subgroup_patterns):
            for group2 in sorted(node.groups):
                if pattern.search(group2.name) is not None and group2 != group:
                    yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(
                        group.name, group2.name)

        if group in node._attributes.get('groups', set()):
            yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                node.name, group.name)
        elif node in group._nodes_from_members:
            yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                group.name, node.name)
        else:
            for pattern in sorted(group._member_patterns):
                if pattern.search(node.name) is not None:
                    yield "\"{}\" -> \"{}\" [color=\"#714D99\",penwidth=2]".format(
                        group.name, node.name)
    yield "}"


def remove_items_not_contributing_to_loop(items):
    """
    We have found a loop. By definition, each item in a loop
    must have at least one incoming and one outgoing dependency.

    We can therefore remove all items without either incoming or
    outgoing dependencies to make the loop more apparent.
    """
    items_with_no_incoming_or_outgoing_deps = set()
    for item in items:
        if not item._deps:
            items_with_no_incoming_or_outgoing_deps.add(item)
        else:
            if item in item._deps:
                continue
            for other_item in items:
                if item == other_item:
                    continue
                if item in other_item._deps:
                    break
            else:
                items_with_no_incoming_or_outgoing_deps.add(item)

    filtered_items = list(filter(
        lambda item: item not in items_with_no_incoming_or_outgoing_deps,
        items,
    ))

    if len(filtered_items) == len(items):
        # nothing happened, end recursion
        return filtered_items
    else:
        # we removed something, this might free up other items we can
        # catch in a second pass
        return remove_items_not_contributing_to_loop(filtered_items)



from datetime import datetime
from getpass import getuser
import json
from os import environ
from shlex import quote
from socket import gethostname
from time import time

from .exceptions import NodeLockedException, NoSuchNode, RemoteException
from .utils import cached_property, tempfile
from .utils.text import (
    blue,
    bold,
    format_duration,
    format_timestamp,
    mark_for_translation as _,
    parse_duration,
    red,
    wrap_question,
)
from .utils.ui import io


def identity():
    return environ.get('BW_IDENTITY', "{}@{}".format(
        getuser(),
        gethostname(),
    ))


class NodeLock:
    def __init__(self, node, interactive=False, ignore=False):
        self.node = node
        self.ignore = ignore
        self.interactive = interactive
        self.locking_node = _get_locking_node(node)

    def __enter__(self):
        if self.locking_node.os not in self.locking_node.OS_FAMILY_UNIX:
            # no locking required/possible
            return self
        with tempfile() as local_path:
            self.locking_node.run("mkdir -p " + quote(self.locking_node.lock_dir))
            if not self.ignore:
                with io.job(_("{node}  checking hard lock status").format(node=bold(self.node.name))):
                    result = self.locking_node.run("mkdir " + quote(self._hard_lock_dir()), may_fail=True)
                    if result.return_code != 0:
                        info = self._get_hard_lock_info(local_path)
                        expired = False
                        try:
                            d = info['date']
                        except KeyError:
                            info['date'] = _("<unknown>")
                            info['duration'] = _("<unknown>")
                        else:
                            duration = datetime.now() - datetime.fromtimestamp(d)
                            info['date'] = format_timestamp(d)
                            info['duration'] = format_duration(duration)
                            if duration > parse_duration(environ.get('BW_HARDLOCK_EXPIRY', "8h")):
                                expired = True
                                io.debug("ignoring expired hard lock on {}".format(self.node.name))
                        if 'user' not in info:
                            info['user'] = _("<unknown>")
                        if expired or self.ignore or (self.interactive and io.ask(
                            self._warning_message_hard(info),
                            False,
                            epilogue=blue("?") + " " + bold(self.node.name),
                        )):
                            pass
                        else:
                            raise NodeLockedException(info)

            with io.job(_("{node}  uploading lock file").format(node=bold(self.node.name))):
                if self.ignore:
                    self.locking_node.run("mkdir -p " + quote(self._hard_lock_dir()))
                with open(local_path, 'w') as f:
                    f.write(json.dumps({
                        'date': time(),
                        'user': identity(),
                    }))
                self.locking_node.upload(local_path, self._hard_lock_file())

        return self

    def __exit__(self, type, value, traceback):
        if self.locking_node.os not in self.locking_node.OS_FAMILY_UNIX:
            # no locking required/possible
            return
        with io.job(_("{node}  removing hard lock").format(node=bold(self.node.name))):
            result = self.locking_node.run("rm -R {}".format(quote(self._hard_lock_dir())), may_fail=True)

        if result.return_code != 0:
            io.stderr(_("{x} {node}  could not release hard lock").format(
                node=bold(self.node.name),
                x=red("!"),
            ))

    def _get_hard_lock_info(self, local_path):
        try:
            self.locking_node.download(self._hard_lock_file(), local_path)
            with open(local_path, 'r') as fp:
                return json.load(fp)
        except (RemoteException, ValueError):
                io.stderr(_(
                    "{x} {node_bold}  corrupted hard lock: "
                    "unable to read or parse lock file contents "
                    "(clear it with `bw run {node} 'rm -Rf {path}'`)"
                ).format(
                    node_bold=bold(self.locking_node.name),
                    node=self.locking_node.name,
                    path=self._hard_lock_dir(),
                    x=red("!"),
                ))
                return {}

    def _hard_lock_dir(self):
        return self.locking_node.lock_dir + "/hard-" + quote(self.node.name)

    def _hard_lock_file(self):
        return self._hard_lock_dir() + "/info"

    def _warning_message_hard(self, info):
        return wrap_question(
            red(_("NODE LOCKED")),
            _(
                "Looks like somebody is currently using BundleWrap on this node.\n"
                "You should let them finish or override the lock if it has gone stale.\n"
                "\n"
                "locked by  {user}\n"
                "    since  {date} ({duration} ago)"
            ).format(
                user=bold(info['user']),
                date=info['date'],
                duration=info['duration'],
            ),
            bold(_("Override lock?")),
            prefix="{x} {node} ".format(node=bold(self.node.name), x=blue("?")),
        )

    @cached_property
    def soft_locks(self):
        return softlock_list(self.node)

    @cached_property
    def my_soft_locks(self):
        for lock in self.soft_locks:
            if lock['user'] == identity():
                yield lock

    @cached_property
    def other_peoples_soft_locks(self):
        for lock in self.soft_locks:
            if lock['user'] != identity():
                yield lock


def _get_locking_node(node):
    if node.locking_node is not None:
        try:
            return node.repo.get_node(node.locking_node)
        except NoSuchNode:
            raise Exception("Invalid locking_node {} for {}".format(
                node.locking_node,
                node.name,
            ))
    else:
        return node


def _soft_lock_dir(node_name, locking_node):
    return locking_node.lock_dir + "/soft-" + quote(node_name)


def _soft_lock_file(node_name, locking_node, lock_id):
    return _soft_lock_dir(node_name, locking_node) + "/" + lock_id


def softlock_add(node, lock_id, comment="", expiry="8h", item_selectors=None):
    locking_node = _get_locking_node(node)
    assert locking_node.os in locking_node.OS_FAMILY_UNIX
    if "\n" in comment:
        raise ValueError(_("Lock comments must not contain any newlines"))
    if not item_selectors:
        item_selectors = ["*"]

    expiry_timedelta = parse_duration(expiry)
    now = time()
    expiry_timestamp = now + expiry_timedelta.days * 86400 + expiry_timedelta.seconds

    content = json.dumps({
        'comment': comment,
        'date': now,
        'expiry': expiry_timestamp,
        'id': lock_id,
        'items': item_selectors,
        'user': identity(),
    }, indent=None, sort_keys=True)

    with tempfile() as local_path:
        with open(local_path, 'w') as f:
            f.write(content + "\n")
        locking_node.run("mkdir -p " + quote(_soft_lock_dir(node.name, locking_node)))
        locking_node.upload(local_path, _soft_lock_file(node.name, locking_node, lock_id), mode='0644')

    node.repo.hooks.lock_add(node.repo, node, lock_id, item_selectors, expiry_timestamp, comment)

    return lock_id


def softlock_list(node):
    locking_node = _get_locking_node(node)
    if locking_node.os not in locking_node.OS_FAMILY_UNIX:
        return []
    with io.job(_("{}  checking soft locks").format(bold(node.name))):
        cat = locking_node.run("cat {}".format(_soft_lock_file(node.name, locking_node, "*")), may_fail=True)
        if cat.return_code != 0:
            return []
        result = []
        for line in cat.stdout.decode('utf-8').strip().split("\n"):
            try:
                result.append(json.loads(line.strip()))
            except json.decoder.JSONDecodeError:
                io.stderr(_(
                    "{x} {node}  unable to parse soft lock file contents, ignoring: {line}"
                ).format(
                    x=red("!"),
                    node=bold(node.name),
                    line=line.strip(),
                ))
        for lock in result[:]:
            if lock['expiry'] < time():
                io.debug(_("removing expired soft lock {id} from node {node}").format(
                    id=lock['id'],
                    node=node.name,
                ))
                softlock_remove(node, lock['id'])
                result.remove(lock)
        return result


def softlock_remove(node, lock_id):
    locking_node = _get_locking_node(node)
    assert locking_node.os in locking_node.OS_FAMILY_UNIX
    io.debug(_("removing soft lock {id} from node {node}").format(
        id=lock_id,
        node=node.name,
    ))
    locking_node.run("rm {}".format(_soft_lock_file(node.name, locking_node, lock_id)))
    node.repo.hooks.lock_remove(node.repo, node, lock_id)



from concurrent.futures import ThreadPoolExecutor, wait, FIRST_COMPLETED
from datetime import datetime
from random import randint
from sys import exit
from traceback import format_tb

from .utils.text import mark_for_translation as _
from .utils.ui import io, QUIT_EVENT

JOIN_TIMEOUT = 5  # seconds


class WorkerPool:
    """
    Manages a bunch of worker threads.
    """
    def __init__(
        self,
        tasks_available,
        next_task,
        handle_result=None,
        handle_exception=None,
        cleanup=None,
        pool_id=None,
        workers=4,
    ):
        if workers < 1:
            raise ValueError(_("at least one worker is required"))

        self.tasks_available = tasks_available
        self.next_task = next_task
        self.handle_result = handle_result
        self.handle_exception = handle_exception
        self.cleanup = cleanup

        self.number_of_workers = workers
        self.idle_workers = set(range(self.number_of_workers))

        self.pool_id = "unnamed_pool_{}".format(randint(1, 99999)) if pool_id is None else pool_id
        self.pending_futures = {}

    def _get_result(self):
        """
        Blocks until a result from a worker is received.
        """
        io.debug(_("worker pool {pool} waiting for next task to complete").format(
            pool=self.pool_id,
        ))
        completed, pending = wait(
            self.pending_futures.keys(),
            return_when=FIRST_COMPLETED,
        )
        future = completed.pop()

        start_time = self.pending_futures[future]['start_time']
        task_id = self.pending_futures[future]['task_id']
        worker_id = self.pending_futures[future]['worker_id']

        del self.pending_futures[future]
        self.idle_workers.add(worker_id)

        exception = future.exception()
        if exception:
            io.debug(_(
                "exception raised while executing task {task} on worker #{worker} "
                "of worker pool {pool}"
            ).format(
                pool=self.pool_id,
                task=task_id,
                worker=worker_id,
            ))
            exception.__task_id = task_id
            raise exception
        else:
            io.debug(_(
                "worker pool {pool} delivering result of {task} on worker #{worker}"
            ).format(
                pool=self.pool_id,
                task=task_id,
                worker=worker_id,
            ))
            return (task_id, future.result(), datetime.now() - start_time)

    def start_task(self, target=None, task_id=None, args=None, kwargs=None):
        """
        target      any callable (includes bound methods)
        task_id     something to remember this worker by
        args        list of positional arguments passed to target
        kwargs      dictionary of keyword arguments passed to target
        """
        if args is None:
            args = []
        else:
            args = list(args)
        if kwargs is None:
            kwargs = {}

        task_id = "unnamed_task_{}".format(randint(1, 99999)) if task_id is None else task_id
        worker_id = self.idle_workers.pop()

        io.debug(_("worker pool {pool} is starting task {task} on worker #{worker}").format(
            pool=self.pool_id,
            task=task_id,
            worker=worker_id,
        ))
        self.pending_futures[self.executor.submit(target, *args, **kwargs)] = {
            'start_time': datetime.now(),
            'task_id': task_id,
            'worker_id': worker_id,
        }

    def run(self):
        io.debug(_("spinning up worker pool {pool}").format(pool=self.pool_id))
        processed_results = []
        exit_code = None
        self.executor = ThreadPoolExecutor(max_workers=self.number_of_workers)
        try:
            while (
                (self.tasks_available() and not QUIT_EVENT.is_set()) or
                self.workers_are_running
            ):
                while (
                    self.tasks_available() and
                    self.workers_are_available and
                    not QUIT_EVENT.is_set()
                ):
                    task = self.next_task()
                    if task is not None:
                        self.start_task(**task)

                if self.workers_are_running:
                    try:
                        result = self._get_result()
                    except SystemExit as exc:
                        if exit_code is None:
                            # Don't overwrite exit code if it has already been set.
                            # This may be a worker exiting with 0 only because
                            # a previous worker raised SystemExit with 1.
                            # We must preserve that original exit code.
                            exit_code = exc.code
                        # just make sure QUIT_EVENT is set and continue
                        # waiting for pending results
                        QUIT_EVENT.set()
                    except Exception as exc:
                        traceback = "".join(format_tb(exc.__traceback__))
                        if self.handle_exception is None:
                            raise exc
                        else:
                            processed_results.append(
                                self.handle_exception(exc.__task_id, exc, traceback)
                            )
                    else:
                        if self.handle_result is not None:
                            processed_results.append(self.handle_result(*result))
            if QUIT_EVENT.is_set():
                # we have reaped all our workers, let's stop this thread
                # before it does anything else
                exit(0 if exit_code is None else exit_code)
            return processed_results
        finally:
            io.debug(_("shutting down worker pool {pool}").format(pool=self.pool_id))
            if self.cleanup:
                self.cleanup()
            self.executor.shutdown()
            io.debug(_("worker pool {pool} has been shut down").format(pool=self.pool_id))

    @property
    def workers_are_available(self):
        return bool(self.idle_workers)

    @property
    def workers_are_running(self):
        return bool(self.pending_futures)



from os import mkdir, readlink, symlink
from os.path import join

from bundlewrap.utils.testing import host_os, make_repo, run


def test_create(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'symlinks': {
                        join(str(tmpdir), "foo"): {
                            'target': "/dev/null",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert readlink(join(str(tmpdir), "foo")) == "/dev/null"


def test_fix(tmpdir):
    symlink(join(str(tmpdir), "bar"), join(str(tmpdir), "foo"))
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'symlinks': {
                        join(str(tmpdir), "foo"): {
                            'target': "/dev/null",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert readlink(join(str(tmpdir), "foo")) == "/dev/null"


def test_fix_dir(tmpdir):
    mkdir(join(str(tmpdir), "foo"))
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'symlinks': {
                        join(str(tmpdir), "foo"): {
                            'target': "/dev/null",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert readlink(join(str(tmpdir), "foo")) == "/dev/null"


def test_fix_dir_target(tmpdir):
    mkdir(join(str(tmpdir), "dir1"))
    mkdir(join(str(tmpdir), "dir2"))
    symlink(join(str(tmpdir), "dir1"), join(str(tmpdir), "link"))
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'symlinks': {
                        join(str(tmpdir), "link"): {
                            'target': join(str(tmpdir), "dir2"),
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert readlink(join(str(tmpdir), "link")) == join(str(tmpdir), "dir2")



from bundlewrap.utils.table import ROW_SEPARATOR, render_table


def test_render_table():
    assert "\n".join(render_table([
        ["head1", "h2"],
        ROW_SEPARATOR,
        ["1", "2"]
    ], alignments={0: 'right'})) == """
╭───────┬────╮
│ head1 │ h2 │
├───────┼────┤
│     1 │ 2  │
╰───────┴────╯
    """.strip()



from bundlewrap.utils.testing import make_repo, run


def test_empty(tmpdir):
    make_repo(tmpdir)
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stdout == b""
    assert stderr == b""
    assert rcode == 0


def test_single(tmpdir):
    make_repo(tmpdir, nodes={"node1": {}})
    stdout, stderr, rcode = run("bw nodes", path=str(tmpdir))
    assert stdout == b"node1\n"
    assert stderr == b""
    assert rcode == 0


def test_hostname(tmpdir):
    make_repo(
        tmpdir,
        groups={"all": {'member_patterns': {r".*"}}},
        nodes={"node1": {'hostname': "node1.example.com"}},
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw nodes all -a hostname | cut -f 2", path=str(tmpdir))
    assert stdout == b"node1.example.com\n"
    assert stderr == b""
    assert rcode == 0


def test_bundles(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "bundle1": {},
            "bundle2": {},
        },
        groups={"all": {'member_patterns': {r".*"}}},
        nodes={
            "node1": {'bundles': ["bundle1", "bundle2"]},
            "node2": {'bundles': ["bundle2"]},
        },
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw nodes all -a bundles | grep node1 | cut -f 2", path=str(tmpdir))
    assert stdout.decode().strip().split("\n") == ["bundle1", "bundle2"]
    assert stderr == b""
    assert rcode == 0


def test_bundles_via_group(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "bundle1": {},
            "bundle2": {},
            "bundle3": {},
        },
        groups={
            "group1": {
                'bundles': {"bundle2"},
                'subgroups': {"group2"},
            },
            "group2": {
                'bundles': {"bundle3"},
            }
        },
        nodes={
            "node1": {
                'bundles': {"bundle1"},
                'groups': {"group2"},
            },
        },
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw nodes node1 -a bundles | cut -f 2", path=str(tmpdir))
    assert stdout.decode().strip().split("\n") == ["bundle1", "bundle2", "bundle3"]
    assert stderr == b""
    assert rcode == 0



"""
Note that modules in this package have to use absolute imports because
Repository.item_classes loads them as files.
"""
from copy import copy
from datetime import datetime
from inspect import cleandoc
from os.path import join
from textwrap import TextWrapper

from bundlewrap.exceptions import BundleError, ItemDependencyError, FaultUnavailable
from bundlewrap.utils import cached_property
from bundlewrap.utils.dicts import diff_keys, diff_value, hash_statedict, validate_statedict
from bundlewrap.utils.text import force_text, mark_for_translation as _
from bundlewrap.utils.text import blue, bold, italic, wrap_question
from bundlewrap.utils.ui import io
from bundlewrap.operations import run_local


BUILTIN_ITEM_ATTRIBUTES = {
    'cascade_skip': None,
    'comment': None,
    'needed_by': set(),
    'needs': set(),
    'preceded_by': set(),
    'precedes': set(),
    'error_on_missing_fault': False,
    'tags': set(),
    'triggered': False,
    'triggered_by': set(),
    'triggers': set(),
    'unless': "",
    'when_creating': {},
}

wrapper = TextWrapper(
    break_long_words=False,
    break_on_hyphens=False,
    expand_tabs=False,
    replace_whitespace=False,
)


def format_comment(comment):
    result = "\n\n"
    for line in wrapper.wrap(cleandoc(comment)):
        for inlineline in line.split("\n"):
            result += "{} {}\n".format(bold("#"), italic(inlineline))
    return result


class ItemStatus:
    """
    Holds information on a particular Item such as whether it needs
    fixing and what's broken.
    """

    def __init__(self, cdict, sdict):
        self.cdict = cdict
        self.sdict = sdict
        self.keys_to_fix = []
        self.must_be_deleted = (self.sdict is not None and self.cdict is None)
        self.must_be_created = (self.cdict is not None and self.sdict is None)
        if not self.must_be_deleted and not self.must_be_created:
            self.keys_to_fix = diff_keys(cdict, sdict)

    def __repr__(self):
        return "<ItemStatus correct:{}>".format(self.correct)

    @property
    def correct(self):
        return not self.must_be_deleted and not self.must_be_created and not bool(self.keys_to_fix)


def make_normalize(attribute_default):
    """
    This is to ensure you can pass filter() results and such in place of
    lists and have them converted to the proper type automatically.
    """
    if type(attribute_default) in (dict, list, set, tuple):
        def normalize(attribute_value):
            if attribute_value is None:
                return attribute_value
            else:
                return type(attribute_default)(attribute_value)
        return normalize
    else:
        return copy


class Item:
    """
    A single piece of configuration (e.g. a file, a package, a service).
    """
    BUNDLE_ATTRIBUTE_NAME = None
    ITEM_ATTRIBUTES = {}
    ITEM_TYPE_NAME = None
    REQUIRED_ATTRIBUTES = []
    SKIP_REASON_CMDLINE = 1
    SKIP_REASON_DEP_FAILED = 2
    SKIP_REASON_FAULT_UNAVAILABLE = 3
    SKIP_REASON_INTERACTIVE = 4
    SKIP_REASON_INTERACTIVE_ONLY = 5
    SKIP_REASON_NO_TRIGGER = 6
    SKIP_REASON_SOFTLOCK = 7
    SKIP_REASON_UNLESS = 8
    SKIP_REASON_DEP_SKIPPED = 9
    SKIP_REASON_DESC = {
        SKIP_REASON_CMDLINE: _("cmdline"),
        SKIP_REASON_DEP_FAILED: _("dependency failed"),
        SKIP_REASON_FAULT_UNAVAILABLE: _("Fault unavailable"),
        SKIP_REASON_INTERACTIVE: _("declined interactively"),
        SKIP_REASON_INTERACTIVE_ONLY: _("interactive only"),
        SKIP_REASON_NO_TRIGGER: _("not triggered"),
        SKIP_REASON_SOFTLOCK: _("soft locked"),
        SKIP_REASON_UNLESS: _("unless"),
        SKIP_REASON_DEP_SKIPPED: _("dependency skipped"),
    }
    STATUS_OK = 1
    STATUS_FIXED = 2
    STATUS_FAILED = 3
    STATUS_SKIPPED = 4
    STATUS_ACTION_SUCCEEDED = 5
    WHEN_CREATING_ATTRIBUTES = {}

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        """
        Return a list of item types that cannot be applied in parallel
        with this item type.
        """
        return []

    def __init__(
        self,
        bundle,
        name,
        attributes,
        skip_validation=False,
        skip_name_validation=False,
    ):
        self.attributes = {}
        self.bundle = bundle
        self.has_been_triggered = False
        self.item_dir = join(bundle.bundle_dir, self.BUNDLE_ATTRIBUTE_NAME)
        self.item_data_dir = join(bundle.bundle_data_dir, self.BUNDLE_ATTRIBUTE_NAME)
        self.name = name
        self.node = bundle.node
        self.when_creating = {}
        self._command_results = []
        self._faults_missing_for_attributes = set()
        self._precedes_items = set()

        if not skip_validation:
            if not skip_name_validation:
                self._validate_name(bundle, name)
                self.validate_name(bundle, name)
            self._validate_attribute_names(bundle, self.id, attributes)
            self._validate_required_attributes(bundle, self.id, attributes)
            self.validate_attributes(bundle, self.id, attributes)

        try:
            attributes = self.patch_attributes(attributes)
        except FaultUnavailable:
            self._faults_missing_for_attributes.add(_("unknown"))

        for attribute_name, attribute_default in BUILTIN_ITEM_ATTRIBUTES.items():
            normalize = make_normalize(attribute_default)
            try:
                setattr(self, attribute_name, force_text(normalize(attributes.get(
                    attribute_name,
                    copy(attribute_default),
                ))))
            except FaultUnavailable:
                self._faults_missing_for_attributes.add(attribute_name)
                setattr(self, attribute_name, BUILTIN_ITEM_ATTRIBUTES[attribute_name])

        for attribute_name, attribute_default in self.ITEM_ATTRIBUTES.items():
            if attribute_name not in BUILTIN_ITEM_ATTRIBUTES:
                normalize = make_normalize(attribute_default)
                try:
                    self.attributes[attribute_name] = force_text(normalize(attributes.get(
                        attribute_name,
                        copy(attribute_default),
                    )))
                except FaultUnavailable:
                    self._faults_missing_for_attributes.add(attribute_name)

        for attribute_name, attribute_default in self.WHEN_CREATING_ATTRIBUTES.items():
            normalize = make_normalize(attribute_default)
            try:
                self.when_creating[attribute_name] = force_text(normalize(
                    attributes.get('when_creating', {}).get(
                        attribute_name,
                        copy(attribute_default),
                    )
                ))
            except FaultUnavailable:
                self._faults_missing_for_attributes.add('when_creating/' + attribute_name)

        if self.cascade_skip is None:
            self.cascade_skip = not (self.unless or self.triggered)

        if self.id in self.triggers:
            raise BundleError(_(
                "item {item} in bundle '{bundle}' can't trigger itself"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
            ))

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        return self.id < other.id

    def __str__(self):
        return self.id

    def __repr__(self):
        return "<Item {}>".format(self.id)

    def _check_bundle_collisions(self, items):
        for item in items:
            if item == self:
                continue
            if item.id == self.id:
                raise BundleError(_(
                    "duplicate definition of {item} in bundles '{bundle1}' and '{bundle2}'"
                ).format(
                    item=item.id,
                    bundle1=item.bundle.name,
                    bundle2=self.bundle.name,
                ))

    def _check_loopback_dependency(self):
        """
        Alerts the user if they have an item depend on itself.
        """
        if self.id in self.needs or self.id in self.needed_by:
            raise ItemDependencyError(_(
                "'{item}' in bundle '{bundle}' on node '{node}' cannot depend on itself"
            ).format(
                item=self.id,
                bundle=self.bundle.name,
                node=self.node.name,
            ))

    @cached_property
    def cached_cdict(self):
        if self._faults_missing_for_attributes:
            self._raise_for_faults()

        cdict = self.cdict()
        try:
            validate_statedict(cdict)
        except ValueError as e:
            raise ValueError(_(
                "{item} from bundle '{bundle}' returned invalid cdict: {msg}"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
                msg=repr(e),
            ))
        return cdict

    @cached_property
    def cached_sdict(self):
        status = self.sdict()
        try:
            validate_statedict(status)
        except ValueError as e:
            raise ValueError(_(
                "{item} from bundle '{bundle}' returned invalid status: {msg}"
            ).format(
                bundle=self.bundle.name,
                item=self.id,
                msg=repr(e),
            ))
        return status

    @cached_property
    def cached_status(self):
        return self.get_status()

    @cached_property
    def cached_unless_result(self):
        """
        Returns True if 'unless' wants to skip this item.
        """
        if self.unless and (self.ITEM_TYPE_NAME == 'action' or not self.cached_status.correct):
            unless_result = self.node.run(self.unless, may_fail=True)
            return unless_result.return_code == 0
        else:
            return False

    def _triggers_preceding_items(self, interactive=False):
        """
        Preceding items will execute this to figure out if they're
        triggered.
        """
        if self.cached_unless_result:
            # 'unless' says we don't need to run
            return False
        if self.ITEM_TYPE_NAME == 'action':
            # so we have an action where 'unless' says it must be run
            # but the 'interactive' attribute might still override that
            if self.attributes['interactive'] and not interactive:
                return False
            else:
                return True
        return not self.cached_status.correct

    def _raise_for_faults(self):
        raise FaultUnavailable(_(
            "{item} on {node} is missing faults "
            "for these attributes: {attrs} "
            "(most of the time this means you're missing "
            "a required key in your .secrets.cfg)"
        ).format(
            attrs=", ".join(sorted(self._faults_missing_for_attributes)),
            item=self.id,
            node=self.node.name,
        ))

    def _skip_with_soft_locks(self, mine, others):
        """
        Returns True/False depending on whether the item should be
        skipped based on the given set of locks.
        """
        for lock in mine:
            if self.covered_by_autoskip_selector(lock['items']):
                io.debug(_("{item} on {node} whitelisted by lock {lock}").format(
                    item=self.id,
                    lock=lock['id'],
                    node=self.node.name,
                ))
                return False
        for lock in others:
            if self.covered_by_autoskip_selector(lock['items']):
                io.debug(_("{item} on {node} blacklisted by lock {lock}").format(
                    item=self.id,
                    lock=lock['id'],
                    node=self.node.name,
                ))
                return True
        return False

    def _test(self):
        with io.job(_("{node}  {bundle}  {item}").format(
            bundle=bold(self.bundle.name),
            item=self.id,
            node=bold(self.node.name),
        )):
            if self._faults_missing_for_attributes:
                self._raise_for_faults()
            return self.test()

    @classmethod
    def _validate_attribute_names(cls, bundle, item_id, attributes):
        if not isinstance(attributes, dict):
            raise BundleError(_(
                "invalid item '{item}' in bundle '{bundle}': not a dict"
            ).format(
                item=item_id,
                bundle=bundle.name,
            ))
        invalid_attributes = set(attributes.keys()).difference(
            set(cls.ITEM_ATTRIBUTES.keys()).union(
                set(BUILTIN_ITEM_ATTRIBUTES.keys())
            ),
        )
        if invalid_attributes:
            raise BundleError(_(
                "invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
            ).format(
                item=item_id,
                bundle=bundle.name,
                attrs=", ".join(invalid_attributes),
            ))

        invalid_attributes = set(attributes.get('when_creating', {}).keys()).difference(
            set(cls.WHEN_CREATING_ATTRIBUTES.keys())
        )
        if invalid_attributes:
            raise BundleError(_(
                "invalid when_creating attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
            ).format(
                item=item_id,
                bundle=bundle.name,
                attrs=", ".join(invalid_attributes),
            ))

    @classmethod
    def _validate_name(cls, bundle, name):
        if ":" in name:
            raise BundleError(_(
                "invalid name for {type} in bundle '{bundle}': {name} (must not contain colon)"
            ).format(
                bundle=bundle.name,
                name=name,
                type=cls.ITEM_TYPE_NAME,
            ))

    @classmethod
    def _validate_required_attributes(cls, bundle, item_id, attributes):
        missing = []
        for attrname in cls.REQUIRED_ATTRIBUTES:
            if attrname not in attributes:
                missing.append(attrname)
        if missing:
            raise BundleError(_(
                "{item} in bundle '{bundle}' missing required attribute(s): {attrs}"
            ).format(
                item=item_id,
                bundle=bundle.name,
                attrs=", ".join(missing),
            ))

    def apply(
        self,
        autoskip_selector=[],
        autoonly_selector=[],
        my_soft_locks=(),
        other_peoples_soft_locks=(),
        interactive=False,
        interactive_default=True,
        show_diff=True,
    ):
        self.node.repo.hooks.item_apply_start(
            self.node.repo,
            self.node,
            self,
        )
        status_code = None
        status_before = None
        status_after = None
        details = None
        start_time = datetime.now()

        if not self.covered_by_autoonly_selector(autoonly_selector):
            io.debug(_(
                "autoonly does not match {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_CMDLINE

        if self.covered_by_autoskip_selector(autoskip_selector):
            io.debug(_(
                "autoskip matches {item} on {node}"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_CMDLINE

        if self._skip_with_soft_locks(my_soft_locks, other_peoples_soft_locks):
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_SOFTLOCK

        for item in self._precedes_items:
            if item._triggers_preceding_items(interactive=interactive):
                io.debug(_(
                    "preceding item {item} on {node} has been triggered by {other_item}"
                ).format(item=self.id, node=self.node.name, other_item=item.id))
                self.has_been_triggered = True
                break
            else:
                io.debug(_(
                    "preceding item {item} on {node} has NOT been triggered by {other_item}"
                ).format(item=self.id, node=self.node.name, other_item=item.id))

        if self.triggered and not self.has_been_triggered and status_code is None:
            io.debug(_(
                "skipping {item} on {node} because it wasn't triggered"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_NO_TRIGGER

        if status_code is None and self.cached_unless_result and status_code is None:
            io.debug(_(
                "'unless' for {item} on {node} succeeded, not fixing"
            ).format(item=self.id, node=self.node.name))
            status_code = self.STATUS_SKIPPED
            details = self.SKIP_REASON_UNLESS

        if self._faults_missing_for_attributes and status_code is None:
            if self.error_on_missing_fault:
                self._raise_for_faults()
            else:
                io.debug(_(
                    "skipping {item} on {node} because it is missing faults "
                    "for these attributes: {attrs} "
                    "(most of the time this means you're missing "
                    "a required key in your .secrets.cfg)"
                ).format(
                    attrs=", ".join(sorted(self._faults_missing_for_attributes)),
                    item=self.id,
                    node=self.node.name,
                ))
                status_code = self.STATUS_SKIPPED
                details = self.SKIP_REASON_FAULT_UNAVAILABLE

        if status_code is None:
            try:
                status_before = self.cached_status
            except FaultUnavailable:
                if self.error_on_missing_fault:
                    self._raise_for_faults()
                else:
                    io.debug(_(
                        "skipping {item} on {node} because it is missing Faults "
                        "(most of the time this means you're missing "
                        "a required key in your .secrets.cfg)"
                    ).format(
                        item=self.id,
                        node=self.node.name,
                    ))
                    status_code = self.STATUS_SKIPPED
                    details = self.SKIP_REASON_FAULT_UNAVAILABLE
            else:
                if status_before.correct:
                    status_code = self.STATUS_OK
                elif show_diff or interactive:
                    if status_before.must_be_created:
                        details = self.display_on_create(copy(status_before.cdict))
                    elif status_before.must_be_deleted:
                        details = self.display_on_delete(copy(status_before.sdict))
                    else:
                        details = self.display_dicts(
                            copy(status_before.cdict),
                            copy(status_before.sdict),
                            # TODO remove sorted() in 5.0 to pass a set
                            sorted(copy(status_before.keys_to_fix)),
                        )

        if status_code is None:
            if not interactive:
                with io.job(_("{node}  {bundle}  {item}").format(
                    bundle=bold(self.bundle.name),
                    item=self.id,
                    node=bold(self.node.name),
                )):
                    self.fix(status_before)
            else:
                if status_before.must_be_created:
                    question_text = "\n".join(
                        f"{bold(key)}  {value}"
                        for key, value in sorted(details.items())
                    )
                    prompt = _("Create {}?").format(bold(self.id))
                elif status_before.must_be_deleted:
                    question_text = "\n".join(
                        f"{bold(key)}  {value}"
                        for key, value in sorted(details.items())
                    )
                    prompt = _("Delete {}?").format(bold(self.id))
                else:
                    display_cdict, display_sdict, display_keys_to_fix = details
                    question_text = "\n".join(
                        diff_value(
                            key,
                            display_sdict[key],
                            display_cdict[key],
                        )
                        for key in sorted(display_keys_to_fix)
                    )
                    prompt = _("Fix {}?").format(bold(self.id))
                if self.comment:
                    question_text += format_comment(self.comment)
                question = wrap_question(
                    self.id,
                    question_text,
                    prompt,
                    prefix="{x} {node} ".format(
                        node=bold(self.node.name),
                        x=blue("?"),
                    ),
                )
                answer = io.ask(
                    question,
                    interactive_default,
                    epilogue="{x} {node}".format(
                        node=bold(self.node.name),
                        x=blue("?"),
                    ),
                )
                if answer:
                    with io.job(_("{node}  {bundle}  {item}").format(
                        bundle=bold(self.bundle.name),
                        item=self.id,
                        node=bold(self.node.name),
                    )):
                        self.fix(status_before)
                else:
                    status_code = self.STATUS_SKIPPED
                    details = self.SKIP_REASON_INTERACTIVE

        if status_code is None:
            status_after = self.get_status(cached=False)
            status_code = self.STATUS_FIXED if status_after.correct else self.STATUS_FAILED

        self.node.repo.hooks.item_apply_end(
            self.node.repo,
            self.node,
            self,
            duration=datetime.now() - start_time,
            status_code=status_code,
            status_before=status_before,
            status_after=status_after,
        )
        return (
            status_code,
            details,
            status_before.must_be_created if status_before else None,
            status_before.must_be_deleted if status_before else None,
        )

    def run_local(self, command, **kwargs):
        result = run_local(command, **kwargs)
        self._command_results.append({
            'command': command,
            'result': result,
        })
        return result

    def run(self, command, **kwargs):
        result = self.node.run(command, **kwargs)
        self._command_results.append({
            'command': command,
            'result': result,
        })
        return result

    def cdict(self):
        """
        Return a statedict that describes the target state of this item
        as configured in the repo. Returning `None` instead means that
        the item should not exist.

        MAY be overridden by subclasses.
        """
        return self.attributes

    def covered_by_autoskip_selector(self, autoskip_selector):
        """
        True if this item should be skipped based on the given selector
        string (e.g. "tag:foo,bundle:bar").
        """
        components = [c.strip() for c in autoskip_selector]
        if (
            "*" in components or
            self.id in components or
            "bundle:{}".format(self.bundle.name) in components or
            "{}:".format(self.ITEM_TYPE_NAME) in components
        ):
            return True
        for tag in self.tags:
            if "tag:{}".format(tag) in components:
                return True
        return False

    def covered_by_autoonly_selector(self, autoonly_selector):
        """
        True if this item should be NOT skipped based on the given selector
        string (e.g. "tag:foo,bundle:bar").
        """
        if not autoonly_selector:
            return True
        components = [c.strip() for c in autoonly_selector]
        if (
            self.id in components or
            "bundle:{}".format(self.bundle.name) in components or
            "{}:".format(self.ITEM_TYPE_NAME) in components
        ):
            return True
        for tag in self.tags:
            if "tag:{}".format(tag) in components:
                return True
        for depending_item in self._incoming_deps:
            if (
                depending_item.id in components or
                "bundle:{}".format(depending_item.bundle.name) in components or
                "{}:".format(depending_item.ITEM_TYPE_NAME) in components
            ):
                return True
            for tag in depending_item.tags:
                if "tag:{}".format(tag) in components:
                    return True
        return False

    def fix(self, status):
        """
        This is supposed to actually implement stuff on the target node.

        MUST be overridden by subclasses.
        """
        raise NotImplementedError()

    def get_auto_deps(self, items):
        """
        Return a list of item IDs this item should have dependencies on.

        Be very careful when using this. There are few circumstances
        where this is really necessary. Only use this if you really need
        to examine the actual list of items in order to figure out your
        dependencies.

        MAY be overridden by subclasses.
        """
        return []

    def get_canned_actions(self):
        """
        Return a dictionary of action definitions (mapping action names
        to dicts of action attributes, as in bundles).

        MAY be overridden by subclasses.
        """
        return {}

    def get_status(self, cached=True):
        """
        Returns an ItemStatus instance describing the current status of
        the item on the actual node.
        """
        with io.job(_("{node}  {bundle}  {item}").format(
            bundle=bold(self.bundle.name),
            item=self.id,
            node=bold(self.node.name),
        )):
            if not cached:
                del self._cache['cached_sdict']
            return ItemStatus(self.cached_cdict, self.cached_sdict)

    def hash(self):
        return hash_statedict(self.cached_cdict)

    @property
    def id(self):
        if self.ITEM_TYPE_NAME == 'action' and ":" in self.name:
            # canned actions don't have an "action:" prefix
            return self.name
        return "{}:{}".format(self.ITEM_TYPE_NAME, self.name)

    def verify(self):
        if self.cached_status.must_be_created:
            display = self.display_on_create(copy(self.cached_status.cdict))
        elif self.cached_status.must_be_deleted:
            display = self.display_on_delete(copy(self.cached_status.sdict))
        else:
            display = self.display_dicts(
                copy(self.cached_status.cdict),
                copy(self.cached_status.sdict),
                # TODO remove sorted() in 5.0 to pass a set
                sorted(copy(self.cached_status.keys_to_fix)),
            )
        return self.cached_unless_result, self.cached_status, display

    def display_on_create(self, cdict):
        """
        Given a cdict as implemented above, modify it to better suit
        interactive presentation when an item is created.

        MAY be overridden by subclasses.
        """
        return cdict

    # TODO rename to display_on_fix in 5.0
    def display_dicts(self, cdict, sdict, keys):
        """
        Given cdict and sdict as implemented above, modify them to
        better suit interactive presentation. The keys parameter is a
        list of keys whose values differ between cdict and sdict.

        MAY be overridden by subclasses.
        """
        return (cdict, sdict, keys)

    def display_on_delete(self, sdict):
        """
        Given an sdict as implemented above, modify it to better suit
        interactive presentation when an item is deleted.

        MAY be overridden by subclasses.
        """
        return sdict

    def patch_attributes(self, attributes):
        """
        Allows an item to preprocess the attributes it is initialized
        with. Returns the modified attributes dictionary.

        MAY be overridden by subclasses.
        """
        return attributes

    def preview(self):
        """
        Can return a preview of this item as a Unicode string.
        BundleWrap will NOT add a trailing newline.

        MAY be overridden by subclasses.
        """
        raise NotImplementedError()

    def sdict(self):
        """
        Return a statedict that describes the actual state of this item
        on the node. Returning `None` instead means that the item does
        not exist on the node.

        For the item to validate as correct, the values for all keys in
        self.cdict() have to match this statedict.

        MUST be overridden by subclasses.
        """
        raise NotImplementedError()

    def test(self):
        """
        Used by `bw repo test`. Should do as much as possible to detect
        what would become a runtime error during a `bw apply`. Files
        will attempt to render their templates for example.

        SHOULD be overridden by subclasses
        """
        pass

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        """
        Raises BundleError if something is amiss with the user-specified
        attributes.

        SHOULD be overridden by subclasses.
        """
        pass

    @classmethod
    def validate_name(cls, bundle, name):
        """
        Raise BundleError if the given name is not valid (e.g. contains
        invalid characters for this kind of item.

        MAY be overridden by subclasses.
        """
        pass



from collections import OrderedDict
from sys import version_info

from ..metadata import METADATA_TYPES, deepcopy_metadata, validate_metadata, value_at_key_path
from .dicts import ATOMIC_TYPES, map_dict_keys, merge_dict


UNMERGEABLE = tuple(METADATA_TYPES) + tuple(ATOMIC_TYPES.values())


class Metastack:
    """
    Holds a number of metadata layers. When laid on top of one another,
    these layers form complete metadata for a node. Each layer comes
    from one particular source of metadata: a bundle default, a group,
    the node itself, or a metadata reactor. Metadata reactors are unique
    in their ability to revise their own layer each time they are run.
    """

    def __init__(self):
        self._partitions = (
            # We rely heavily on insertion order in these dicts.
            {} if version_info >= (3, 7) else OrderedDict(),  # node/groups
            {} if version_info >= (3, 7) else OrderedDict(),  # reactors
            {} if version_info >= (3, 7) else OrderedDict(),  # defaults
        )
        self._cached_partitions = {}

    def get(self, path):
        """
        Get the value at the given path, merging all layers together.
        """
        result = None
        undef = True

        for part_index, partition in enumerate(self._partitions):
            # prefer cached partitions if available
            partition = self._cached_partitions.get(part_index, partition)
            for layer in reversed(list(partition.values())):
                try:
                    value = value_at_key_path(layer, path)
                except KeyError:
                    pass
                else:
                    if undef:
                        # First time we see anything. If we can't merge
                        # it anyway, then return early.
                        if isinstance(value, UNMERGEABLE):
                            return value
                        result = {'data': value}
                        undef = False
                    else:
                        result = merge_dict({'data': value}, result)

        if undef:
            raise KeyError('/'.join(path))
        else:
            return deepcopy_metadata(result['data'])

    def as_dict(self, partitions=None):
        final_dict = {}

        if partitions is None:
            partitions = tuple(range(len(self._partitions)))
        else:
            partitions = sorted(partitions)

        for part_index in partitions:
            # prefer cached partitions if available
            partition = self._cached_partitions.get(part_index, self._partitions[part_index])
            for layer in reversed(list(partition.values())):
                final_dict = merge_dict(layer, final_dict)

        return final_dict

    def as_blame(self):
        keymap = map_dict_keys(self.as_dict())
        blame = {}
        for path in keymap:
            for partition in self._partitions:
                for identifier, layer in partition.items():
                    try:
                        value_at_key_path(layer, path)
                    except KeyError:
                        pass
                    else:
                        blame.setdefault(path, []).append(identifier)
        return blame

    def pop_layer(self, partition_index, identifier):
        try:
            return self._partitions[partition_index].pop(identifier)
        except (KeyError, IndexError):
            return {}

    def set_layer(self, partition_index, identifier, new_layer):
        validate_metadata(new_layer)
        self._partitions[partition_index][identifier] = new_layer

    def cache_partition(self, partition_index):
        self._cached_partitions[partition_index] = {
            'merged layers': self.as_dict(partitions=[partition_index]),
        }



from collections.abc import Collection
from os import environ, makedirs
from os.path import dirname, exists, join
from sys import exit

from ..deps import prepare_dependencies
from ..exceptions import FaultUnavailable
from ..items import BUILTIN_ITEM_ATTRIBUTES
from ..utils.cmdline import get_item, get_node
from ..utils.dicts import statedict_to_json
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import bold, green, mark_for_translation as _, red, yellow
from ..utils.ui import io, page_lines


def write_preview(file_item, base_path):
    """
    Writes the content of the given file item to the given path.
    """
    # this might raise an exception, try it before creating anything
    content = file_item.content
    file_path = join(base_path, file_item.name.lstrip("/"))
    dir_path = dirname(file_path)
    if not exists(dir_path):
        makedirs(dir_path)
    with open(file_path, 'wb') as f:
        f.write(content)


def bw_items(repo, args):
    node = get_node(repo, args['node'])
    if args['preview'] and not args['item']:
        io.stderr(_("{x} no ITEM given for preview").format(x=red("!!!")))
        exit(1)
    elif args['file_preview_path']:
        if args['item']:
            io.stderr(_("{x} use --file-preview to preview single files").format(x=red("!!!")))
            exit(1)
        if exists(args['file_preview_path']):
            io.stderr(_(
                "not writing to existing path: {path}"
            ).format(path=args['file_preview_path']))
            exit(1)
        for item in sorted(node.items):
            if not item.id.startswith("file:"):
                continue
            if item.attributes['content_type'] == 'any':
                io.stderr(_(
                    "{x} skipped {filename} (content_type 'any')"
                ).format(x=yellow("»"), filename=bold(item.name)))
                continue
            if item.attributes['content_type'] == 'binary':
                io.stderr(_(
                    "{x} skipped {filename} (content_type 'binary')"
                ).format(x=yellow("»"), filename=bold(item.name)))
                continue
            if item.attributes['delete']:
                io.stderr(_(
                    "{x} skipped {filename} ('delete' attribute set)"
                ).format(x=yellow("»"), filename=bold(item.name)))
                continue
            try:
                write_preview(item, args['file_preview_path'])
            except FaultUnavailable:
                io.stderr(_(
                    "{x} skipped {path} (Fault unavailable)"
                ).format(x=yellow("»"), path=bold(item.name)))
            else:
                io.stdout(_(
                    "{x} wrote {path}"
                ).format(
                    x=green("✓"),
                    path=bold(join(
                        args['file_preview_path'],
                        item.name.lstrip("/"),
                    )),
                ))
    elif args['item']:
        item = get_item(node, args['item'])
        if args['preview']:
            try:
                io.stdout(
                    item.preview(),
                    append_newline=False,
                )
            except NotImplementedError:
                io.stderr(_(
                    "{x} cannot preview {item} on {node} (doesn't support previews)"
                ).format(x=red("!!!"), item=item.id, node=node.name))
                exit(1)
            except ValueError:
                io.stderr(_(
                    "{x} cannot preview {item} on {node} (not available for this item config)"
                ).format(x=red("!!!"), item=item.id, node=node.name))
                exit(1)
            except FaultUnavailable:
                io.stderr(_(
                    "{x} cannot preview {item} on {node} (Fault unavailable)"
                ).format(x=red("!!!"), item=item.id, node=node.name))
                exit(1)
        else:
            if args['show_attrs']:
                prepare_dependencies(node)
                table = [[bold(_("attribute")), bold(_("value"))]]
                for attribute in BUILTIN_ITEM_ATTRIBUTES:
                    table.append(ROW_SEPARATOR)
                    value = getattr(item, attribute)
                    if isinstance(value, Collection):
                        value = sorted(value) or [""]
                        table.append([attribute, value.pop(0)])
                        for element in value:
                            table.append([
                                attribute if environ.get('BW_TABLE_STYLE') == "grep" else "",
                                element,
                            ])
                    else:
                        table.append([attribute, repr(value)])
                page_lines(render_table(table))
            else:
                if args['show_sdict']:
                    statedict = item.sdict()
                else:
                    statedict = item.cdict()
                if statedict is None:
                    io.stdout("REMOVE")
                else:
                    if args['attr']:
                        io.stdout(repr(statedict[args['attr']]))
                    else:
                        io.stdout(statedict_to_json(statedict, pretty=True))
    else:
        for item in sorted(node.items):
            if args['show_repr']:
                io.stdout(repr(item))
            else:
                io.stdout(item.id)



from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import force_text, mark_for_translation as _


def svc_start(node, svcname):
    return node.run("systemctl start -- {}".format(quote(svcname)), may_fail=True)


def svc_running(node, svcname):
    result = node.run(
        "systemctl status -- {}".format(quote(svcname)),
        may_fail=True,
    )
    return result.return_code == 0


def svc_stop(node, svcname):
    return node.run("systemctl stop -- {}".format(quote(svcname)), may_fail=True)


def svc_enable(node, svcname):
    return node.run("systemctl enable -- {}".format(quote(svcname)), may_fail=True)


def svc_enabled(node, svcname):
    result = node.run(
        "systemctl is-enabled -- {}".format(quote(svcname)),
        may_fail=True,
    )
    return (
        result.return_code == 0 and
        force_text(result.stdout).strip() != "runtime-enabled"
    )


def svc_disable(node, svcname):
    return node.run("systemctl disable -- {}".format(quote(svcname)), may_fail=True)

def svc_mask(node, svcname):
    return node.run("systemctl mask -- {}".format(quote(svcname)), may_fail=True)

def svc_masked(node, svcname):
    result = node.run(
        "systemctl is-enabled -- {}".format(quote(svcname)),
        may_fail=True,
    )
    return (
        result.return_code == 1 and
        force_text(result.stdout).strip() == "masked"
    )

def svc_unmask(node, svcname):
    return node.run("systemctl unmask -- {}".format(quote(svcname)), may_fail=True)


class SvcSystemd(Item):
    """
    A service managed by systemd.
    """
    BUNDLE_ATTRIBUTE_NAME = "svc_systemd"
    ITEM_ATTRIBUTES = {
        'enabled': True,
        'running': True,
        'masked': False,
    }
    ITEM_TYPE_NAME = "svc_systemd"

    def __repr__(self):
        return "<SvcSystemd name:{} enabled:{} running:{} masked:{}>".format(
            self.name,
            self.attributes['enabled'],
            self.attributes['running'],
            self.attributes['masked'],
        )

    def cdict(self):
        cdict = {}
        for option, value in self.attributes.items():
            if value is not None:
                cdict[option] = value
        return cdict

    def fix(self, status):
        if 'masked' in status.keys_to_fix:
            if self.attributes['masked']:
                svc_mask(self.node, self.name)
            else:
                svc_unmask(self.node, self.name)

        if 'enabled' in status.keys_to_fix:
            if self.attributes['enabled']:
                svc_enable(self.node, self.name)
            else:
                svc_disable(self.node, self.name)

        if 'running' in status.keys_to_fix:
            if self.attributes['running']:
                svc_start(self.node, self.name)
            else:
                svc_stop(self.node, self.name)

    def get_canned_actions(self):
        return {
            'stop': {
                'command': "systemctl stop -- {}".format(self.name),
                'needed_by': {self.id},
            },
            'restart': {
                'command': "systemctl restart -- {}".format(self.name),
                'needs': {self.id},
            },
            'reload': {
                'command': "systemctl reload -- {}".format(self.name),
                'needs': {
                    # make sure we don't reload and restart simultaneously
                    f"{self.id}:restart",
                    # with only the dep on restart, we might still end
                    # up reloading if the service itself is skipped
                    # because the restart action has cascade_skip False
                    self.id,
                },
            },
        }

    def sdict(self):
        return {
            'enabled': svc_enabled(self.node, self.name),
            'running': svc_running(self.node, self.name),
            'masked': svc_masked(self.node, self.name),
        }

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        for attribute in ('enabled', 'running'):
            if attributes.get(attribute, None) not in (True, False, None):
                raise BundleError(_(
                    "expected boolean or None for '{attribute}' on {item} in bundle '{bundle}'"
                ).format(
                    attribute=attribute,
                    bundle=bundle.name,
                    item=item_id,
                ))



from os.path import exists, join

from bundlewrap.utils.testing import host_os, make_repo, run


def test_skip_bundle(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip bundle:test -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_group(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'groups': {"foo"},
                'os': host_os(),
            },
        },
        groups={
            "foo": {},
        },
    )
    result = run("bw apply --skip group:foo -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_id(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip file:{} -- localhost".format(join(str(tmpdir), "foo")), path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_node(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip node:localhost -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_tag(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                            'tags': ["nope"],
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip tag:nope -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_type(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip file: -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))


def test_skip_trigger(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                            'tags': ["nope"],
                            'triggers': ["file:{}".format(join(str(tmpdir), "bar"))],
                        },
                        join(str(tmpdir), "bar"): {
                            'content': "nope",
                            'triggered': True,
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    result = run("bw apply --skip tag:nope -- localhost", path=str(tmpdir))
    assert result[2] == 0
    assert not exists(join(str(tmpdir), "foo"))
    assert not exists(join(str(tmpdir), "bar"))



from bundlewrap.utils.testing import host_os, make_repo, run


def test_run_ok(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "localhost": {
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw run localhost true", path=str(tmpdir))
    assert rcode == 0
    assert b"localhost\t0" in stdout
    assert stderr == b""


def test_run_fail(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "localhost": {
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw run localhost false", path=str(tmpdir))
    assert rcode == 0
    assert b"localhost\t1" in stdout
    assert stderr == b""



from collections import defaultdict, Counter
from contextlib import suppress
from os import environ
from traceback import TracebackException

from .exceptions import MetadataPersistentKeyError
from .metadata import DoNotRunAgain
from .node import _flatten_group_hierarchy
from .utils import list_starts_with, randomize_order, NO_DEFAULT
from .utils.dicts import extra_paths_in_dict
from .utils.ui import io, QUIT_EVENT
from .utils.metastack import Metastack
from .utils.text import bold, mark_for_translation as _, red


MAX_METADATA_ITERATIONS = int(environ.get("BW_MAX_METADATA_ITERATIONS", "1000"))


class PathSet:
    """
    Collects metadata paths and stores only the highest levels ones.

    >>> s = PathSet()
    >>> s.add(("foo", "bar"))
    >>> s.add(("foo",))
    >>> s
    {"foo"}
    """

    def __init__(self):
        self._paths = set()

    def __iter__(self):
        for path in self._paths:
            yield path

    def __len__(self):
        return len(self._paths)

    def __repr__(self):
        return "<PathSet: {}>".format(repr(self._paths))

    def add(self, new_path):
        if self.covers(new_path):
            return False
        for existing_path in self._paths.copy():
            if list_starts_with(existing_path, new_path):
                self._paths.remove(existing_path)
        self._paths.add(new_path)
        return True

    def covers(self, candidate_path):
        for existing_path in self._paths:
            if list_starts_with(candidate_path, existing_path):
                return True
        return False

    def needs(self, candidate_path):
        for existing_path in self._paths:
            if list_starts_with(existing_path, candidate_path):
                return True
        return False


def reactors_for_paths(available_reactors, required_paths):
    """
    Returns only those available_reactors that might affect the
    required_paths.
    """
    for name, reactor in available_reactors:
        provides = getattr(reactor, '_provides', tuple())
        if provides:
            for path in provides:
                if required_paths.covers(path) or required_paths.needs(path):
                    yield name, reactor
                    break
        else:
            yield name, reactor


class NodeMetadataProxy:
    def __init__(self, metagen, node):
        self._metagen = metagen
        self._node = node
        self._metastack = Metastack()
        self._completed_reactors = set()
        self._requested_paths = PathSet()
        self._satisfied = False  # has this node completed all required reactors?
        self.__relevant_reactors_cache = None

    def __contains__(self, key):
        try:
            self.get(key, _backwards_compatibility_default=False)
        except KeyError:
            return False
        else:
            return True

    def __getitem__(self, key):
        return self.get((key,), _backwards_compatibility_default=False)

    def __iter__(self):
        for key, value in self.get(tuple()).items():
            yield key, value

    @property
    def _relevant_reactors(self):
        """
        All reactors that might provide some of the requested paths.
        """
        if self.__relevant_reactors_cache is None:
            self.__relevant_reactors_cache = set(reactors_for_paths(
                self._node.metadata_reactors,
                self._requested_paths,
            ))
        return self.__relevant_reactors_cache

    @property
    def _pending_reactors(self):
        """
        All reactors that might provide some of the requested paths and
        have not yet been run to completion.
        """
        for reactor in self._relevant_reactors:
            if reactor not in self._completed_reactors:
                yield reactor

    @property
    def blame(self):
        if self._metagen._in_a_reactor:
            raise RuntimeError("cannot call node.metadata.blame from a reactor")
        else:
            return self._metastack.as_blame()

    @property
    def stack(self):
        if self._metagen._in_a_reactor:
            raise RuntimeError("cannot call node.metadata.stack from a reactor")
        else:
            return self._metastack

    def get(self, path, default=NO_DEFAULT, _backwards_compatibility_default=True):
        if (
            default == NO_DEFAULT and
            _backwards_compatibility_default and
            not self._metagen._in_a_reactor and
            "/" not in path
        ):
            # make node.metadata.get('foo') work as if it was still a dict
            # TODO remove in 5.0
            default = None
        if not isinstance(path, (tuple, list)):
            path = tuple(path.split("/"))

        with self._metagen._node_metadata_lock:
            # The lock is required because there are several thread-unsafe things going on here:
            #
            #   self._requested_paths
            #   self._metagen._build_node_metadata
            #   self._metastack
            #
            # It needs to be an RLock because this method will be recursively
            # called from _build_node_metadata (when reactors call node.metadata.get()).
            if self._requested_paths.add(path):
                self._satisfied = False
                self.__relevant_reactors_cache = None
                if self._metagen._in_a_reactor:
                    self._metagen._additional_path_requested = True

            if self._metagen._in_a_reactor:
                self._metagen._partial_metadata_accessed_for.add(self._node.name)
            else:
                self._metagen._build_node_metadata(self._node.name)

            try:
                return self._metastack.get(path)
            except KeyError:
                if default != NO_DEFAULT:
                    return default
                else:
                    raise

    def items(self):
        return self.get(tuple()).items()

    def keys(self):
        return self.get(tuple()).keys()

    def values(self):
        return self.get(tuple()).values()


class _StartOver(Exception):
    """
    Raised when metadata processing needs to start from the top.
    """
    pass


class MetadataGenerator:
    # are we currently executing a reactor?
    _in_a_reactor = False
    # should reactor return values be checked against their declared keys?
    _verify_reactor_provides = False

    def __reset(self):
        # reactors that raised DoNotRunAgain
        self.__do_not_run_again = set()
        # reactors that raised KeyErrors (and which ones)
        self.__keyerrors = {}
        # mapping each node to all nodes that depend on it
        self.__node_deps = defaultdict(set)
        # how often __run_reactors was called for a node
        self.__node_iterations = defaultdict(int)
        # nodes we encountered as a dependency through partial_metadata,
        # but haven't run yet
        self.__nodes_that_never_ran = set()
        # nodes whose dependencies changed and that have to rerun their
        # reactors depending on those nodes
        self.__triggered_nodes = set()
        # nodes we already did initial processing on
        self.__nodes_that_ran_at_least_once = set()
        # how often we called reactors
        self.__reactors_run = 0
        # how often each reactor changed
        self.__reactor_changes = defaultdict(int)
        # set when a reactor requests an additional path
        self._additional_path_requested = False

    def _metadata_proxy_for_node(self, node_name):
        if node_name not in self._node_metadata_proxies:
            self._node_metadata_proxies[node_name] = NodeMetadataProxy(self, self.get_node(node_name))
        return self._node_metadata_proxies[node_name]

    def __run_new_nodes(self):
        try:
            node_name = self.__nodes_that_never_ran.pop()
        except KeyError:
            pass
        else:
            self.__nodes_that_ran_at_least_once.add(node_name)
            self.__initial_run_for_node(node_name)
            raise _StartOver

    def __run_triggered_nodes(self):
        try:
            node_name = self.__triggered_nodes.pop()
        except KeyError:
            pass
        else:
            io.debug(f"triggered metadata run for {node_name}")
            self.__run_reactors(self.get_node(node_name))
            raise _StartOver

    def __run_nodes(self):
        encountered_unsatisfied_node = False
        for proxy in randomize_order(self._node_metadata_proxies.values()):
            node = proxy._node
            io.debug(f"running reactors for {node.name}")
            self.__run_reactors(node)
            if not proxy._satisfied:
                io.debug(f"{node.name} still not satisfied")
                encountered_unsatisfied_node = True
            if self.__nodes_that_never_ran:
                # we have found a new dependency, process it immediately
                # going wide early should be more efficient
                raise _StartOver
        if encountered_unsatisfied_node:
            raise _StartOver

    def _build_node_metadata(self, initial_node_name):
        if self._node_metadata_proxies[initial_node_name]._satisfied:
            return

        self.__reset()
        self.__nodes_that_never_ran.add(initial_node_name)

        while not QUIT_EVENT.is_set():
            jobmsg = _("{b} ({n} nodes, {r} reactors, {e} runs)").format(
                b=bold(_("running metadata reactors")),
                n=len(self.__nodes_that_never_ran) + len(self.__nodes_that_ran_at_least_once),
                r=len(self.__reactor_changes),
                e=self.__reactors_run,
            )
            try:
                with io.job(jobmsg):
                    # Control flow here is a bit iffy. The functions in this block often raise
                    # _StartOver in order to aggressively process new nodes first etc.
                    # Each method represents a distinct stage of metadata processing that checks
                    # for nodes in certain states as described below.

                    # This checks for newly discovered nodes that haven't seen any processing at
                    # all so far. It is important that we run them as early as possible, so their
                    # static metadata becomes available to other nodes and we recursively discover
                    # additional nodes as quickly as possible.
                    self.__run_new_nodes()
                    # At this point, we have run all relevant nodes at least once.

                    # Nodes become "triggered" when they previously looked something up from a
                    # different node and that second node changed. In this method, we try to figure
                    # out if the change on the node we depend on actually has any effect on the
                    # depending node.
                    self.__run_triggered_nodes()

                    # The final step is to make sure nothing changes when we run reactors with
                    # dependencies on other nodes. If anything changes, we need to start over so
                    # local-only reactors on a node can react to changes caused by reactors looking
                    # at other nodes.
                    self.__run_nodes()

                    # If we get here, we're done! All that's left to do is blacklist completed
                    # reactors so they don't get run again if additional metadata is requested.
                    for proxy in self._node_metadata_proxies.values():
                        proxy._completed_reactors.update(
                            proxy._relevant_reactors
                        )
                        if proxy._requested_paths.covers(tuple()):  # full metadata
                            proxy._metastack.cache_partition(1)
                    break

            except _StartOver:
                continue

        if self.__keyerrors and not QUIT_EVENT.is_set():
            msg = _(
                "These metadata reactors raised a KeyError "
                "even after all other reactors were done:"
            )
            for source, exc in sorted(self.__keyerrors.items()):
                node_name, reactor = source
                msg += f"\n\n  {node_name} {reactor}\n\n"
                for line in TracebackException.from_exception(exc).format():
                    msg += "    " + line
            raise MetadataPersistentKeyError(msg)

        io.debug("metadata generation for selected nodes finished")

    def __initial_run_for_node(self, node_name):
        io.debug(f"initial metadata run for {node_name}")
        node = self.get_node(node_name)

        # randomize order to increase chance of exposing clashing defaults
        for defaults_name, defaults in randomize_order(node.metadata_defaults):
            node.metadata._metastack.set_layer(
                2,
                defaults_name,
                defaults,
            )
        node.metadata._metastack.cache_partition(2)

        group_order = _flatten_group_hierarchy(node.groups)
        for group_name in group_order:
            node.metadata._metastack.set_layer(
                0,
                "group:{}".format(group_name),
                self.get_group(group_name)._attributes.get('metadata', {}),
            )

        node.metadata._metastack.set_layer(
            0,
            "node:{}".format(node_name),
            node._attributes.get('metadata', {}),
        )
        node.metadata._metastack.cache_partition(0)

        # run all reactors once to get started  TODO is this necessary?
        self.__run_reactors(node)

    def __check_iteration_count(self, node_name):
        self.__node_iterations[node_name] += 1
        if self.__node_iterations[node_name] > MAX_METADATA_ITERATIONS:
            top_changers = Counter(self.__reactor_changes).most_common(25)
            msg = _(
                "MAX_METADATA_ITERATIONS({m}) exceeded for {node}, "
                "likely an infinite loop between flip-flopping metadata reactors.\n"
                "These are the reactors that changed most often:\n\n"
            ).format(m=MAX_METADATA_ITERATIONS, node=node_name)
            for reactor, count in top_changers:
                msg += f"  {count}\t{reactor[0]}\t{reactor[1]}\n"
            raise RuntimeError(msg)

    def __run_reactors(self, node):
        self.__check_iteration_count(node.name)
        any_reactor_changed = False
        self._additional_path_requested = False

        # TODO ideally, we should run the least-run reactors first
        for reactor_name, reactor in randomize_order(
            self._node_metadata_proxies[node.name]._pending_reactors
        ):
            reactor_changed, deps = self.__run_reactor(node, reactor_name, reactor)
            io.debug(f"{node.name}:{reactor_name} changed={reactor_changed} deps={deps}")
            if reactor_changed:
                any_reactor_changed = True
            for required_node_name in deps:
                if required_node_name not in self.__nodes_that_ran_at_least_once:
                    # we found a node that we didn't need until now
                    self.__nodes_that_never_ran.add(required_node_name)
                # this is so we know the current node needs to be run
                # again if the required node changes
                self.__node_deps[required_node_name].add(node.name)

        if any_reactor_changed:
            # something changed on this node, mark all dependent nodes as unstable
            for required_node_name in self.__node_deps[node.name]:
                io.debug(f"{node.name} triggering metadata rerun on {required_node_name}")
                self.__triggered_nodes.add(required_node_name)

        if not self._additional_path_requested and not any_reactor_changed:
            self._node_metadata_proxies[node.name]._satisfied = True

    def __run_reactor(self, node, reactor_name, reactor):
        if (node.name, reactor_name) in self.__do_not_run_again:
            return False, set()
        self._partial_metadata_accessed_for = set()
        self.__reactors_run += 1
        # make sure the reactor doesn't react to its own output
        old_metadata = node.metadata._metastack.pop_layer(1, reactor_name)
        self._in_a_reactor = True
        try:
            new_metadata = reactor(node.metadata)
        except KeyError as exc:
            self.__keyerrors[(node.name, reactor_name)] = exc
            return False, self._partial_metadata_accessed_for
        except DoNotRunAgain:
            self.__do_not_run_again.add((node.name, reactor_name))
            # clear any previously stored exception
            with suppress(KeyError):
                del self.__keyerrors[(node.name, reactor_name)]
            return False, set()
        except Exception as exc:
            io.stderr(_(
                "{x} Exception while executing metadata reactor "
                "{metaproc} for node {node}:"
            ).format(
                x=red("!!!"),
                metaproc=reactor_name,
                node=node.name,
            ))
            raise exc
        finally:
            self._in_a_reactor = False

        # reactor terminated normally, clear any previously stored exception
        with suppress(KeyError):
            del self.__keyerrors[(node.name, reactor_name)]

        if self._verify_reactor_provides and getattr(reactor, '_provides', None):
            extra_paths = extra_paths_in_dict(new_metadata, reactor._provides)
            if extra_paths:
                raise ValueError(_(
                    "{reactor_name} on {node_name} returned the following key paths, "
                    "but didn't declare them with @metadata_reactor.provides():\n"
                    "{paths}"
                ).format(
                    node_name=node.name,
                    reactor_name=reactor_name,
                    paths="\n".join(["/".join(path) for path in sorted(extra_paths)]),
                ))

        try:
            node.metadata._metastack.set_layer(
                1,
                reactor_name,
                new_metadata,
            )
        except TypeError as exc:
            # TODO catch validation errors better
            io.stderr(_(
                "{x} Exception after executing metadata reactor "
                "{metaproc} for node {node}:"
            ).format(
                x=red("!!!"),
                metaproc=reactor_name,
                node=node.name,
            ))
            raise exc

        changed = old_metadata != new_metadata
        if changed:
            self.__reactor_changes[(node.name, reactor_name)] += 1

        return changed, self._partial_metadata_accessed_for



from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_empty(tmpdir):
    make_repo(tmpdir)
    stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
    assert stdout == b"bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f\n"
    assert stderr == b""


def test_nondeterministic(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content_type': 'mako',
                            'content': "<% import random %>${random.randint(1, 9999)}",
                        },
                    },
                },
            },
        },
    )

    hashes = set()

    for i in range(3):
        stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
        hashes.add(stdout.strip())

    assert len(hashes) > 1


def test_deterministic(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "${node.name}",
                            'group': None,  # BSD has a different default and we don't want to
                                            # deal with that here
                        },
                    },
                },
            },
        },
    )

    hashes = set()

    for i in range(3):
        stdout, stderr, rcode = run("bw hash", path=str(tmpdir))
        hashes.add(stdout.strip())

    assert len(hashes) == 1
    assert hashes.pop() == b"2203e7acc35608bbff471c023b7b7498e5b385d9"


def test_dict(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "yes please",
                            'group': None,  # BSD has a different default and we don't want to
                                            # deal with that here
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -d", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"93e7a2c6e8cdc71fb4df5426bc0d0bb978d84381  node1\n"

    stdout, stderr, rcode = run("bw hash -d node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"59d1a7c79640ccdfd3700ab141698a9389fcd0b7  file:/test\n"

    stdout, stderr, rcode = run("bw hash -d node1 file:/test", path=str(tmpdir))
    assert rcode == 0
    assert stdout == (
        b"content_hash\tc05a36d547e2b1682472f76985018038d1feebc5\n"
        b"mode\t0644\n"
        b"owner\troot\n"
        b"type\tfile\n"
    )


def test_metadata_empty(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {},
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f\n"


def test_metadata_fault(tmpdir):
    make_repo(tmpdir)
    with open(join(str(tmpdir), "nodes.py"), 'w') as f:
        f.write("""
nodes = {
    'node1': {
        'metadata': {'foo': vault.password_for("testing")},
    },
    'node2': {
        'metadata': {'foo': vault.password_for("testing").value},
    },
    'node3': {
        'metadata': {'foo': "faCTT76kagtDuZE5wnoiD1CxhGKmbgiX"},
    },
    'node4': {
        'metadata': {'foo': "something else entirely"},
    },
}
""")
    stdout1, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert stdout1 == b"b60c0959c9c1ff38940d7b6d4121b2162be34fc9\n"
    assert stderr == b""
    assert rcode == 0
    stdout2, stderr, rcode = run("bw hash -m node2", path=str(tmpdir))
    assert stdout2 == stdout1
    assert stderr == b""
    assert rcode == 0
    stdout3, stderr, rcode = run("bw hash -m node3", path=str(tmpdir))
    assert stdout3 == stdout1
    assert stderr == b""
    assert rcode == 0
    stdout4, stderr, rcode = run("bw hash -m node4", path=str(tmpdir))
    assert stdout4 != stdout1
    assert stderr == b""
    assert rcode == 0


def test_metadata_nested_sort(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'nested': {
                        'one': True,
                        'two': False,
                        'three': 3,
                        'four': "four",
                        'five': None,
                    },
                },
            },
            "node2": {
                'metadata': {
                    'nested': {
                        'five': None,
                        'four': "four",
                        'one': True,
                        'three': 3,
                        'two': False,
                    },
                },
            },
        },
    )

    stdout1, stderr, rcode = run("bw hash -m node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout1 == b"d96dc8da8948d0da7924954a657ac960ce7194e9\n"

    stdout2, stderr, rcode = run("bw hash -m node2", path=str(tmpdir))
    assert rcode == 0
    assert stdout1 == stdout2


def test_metadata_repo(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'foo': 47,
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -m", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"8c4a30eaa521c966c678d6e51070f6b3a34b7322\n"


def test_metadata_repo_dict(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'metadata': {
                    'foo': 47,
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw hash -md", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"node1\t223fb72805ecab20f92b463af65896303f997f1c\n"


def test_groups_repo(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -g", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"479c737e191339e5fae20ac8a8903a75f6b91f4d\n"


def test_groups_repo_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"group1\ngroup2\n"


def test_groups(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {'members': ["node1", "node2"]},
            "group2": {'members': ["node3"]},
        },
        nodes={
            "node1": {},
            "node2": {},
            "node3": {},
        },
    )

    stdout, stderr, rcode = run("bw hash -g group1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"59f5a812acd22592b046b20e9afedc1cfcd37c77\n"


def test_groups_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg group1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"node1\nnode2\n"


def test_groups_node(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -g node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"6f4615dc71426549e22df7961bd2b88ba95ad1fc\n"


def test_groups_node_dict(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {},
            "group2": {},
        },
        nodes={
            "node1": {'groups': {"group1"}},
            "node2": {'groups': {"group1"}},
            "node3": {'groups': {"group2"}},
        },
    )

    stdout, stderr, rcode = run("bw hash -dg node1", path=str(tmpdir))
    assert rcode == 0
    assert stdout == b"group1\n"



from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


def svc_start(node, svcname):
    return node.run("/etc/rc.d/{} start".format(quote(svcname)), may_fail=True)


def svc_running(node, svcname):
    result = node.run("/etc/rc.d/{} check".format(quote(svcname)), may_fail=True)
    return "ok" in result.stdout_text


def svc_stop(node, svcname):
    return node.run("/etc/rc.d/{} stop".format(quote(svcname)), may_fail=True)


def svc_enable(node, svcname):
    return node.run("rcctl set {} status on".format(quote(svcname)), may_fail=True)


def svc_enabled(node, svcname):
    result = node.run(
        "rcctl ls on | grep '^{}$'".format(svcname),
        may_fail=True,
    )
    return result.return_code == 0


def svc_disable(node, svcname):
    return node.run("rcctl set {} status off".format(quote(svcname)), may_fail=True)


class SvcOpenBSD(Item):
    """
    A service managed by OpenBSD rc.d.
    """
    BUNDLE_ATTRIBUTE_NAME = "svc_openbsd"
    ITEM_ATTRIBUTES = {
        'running': True,
        'enabled': True
    }
    ITEM_TYPE_NAME = "svc_openbsd"

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        # https://github.com/bundlewrap/bundlewrap/issues/554
        return [cls.ITEM_TYPE_NAME]

    def __repr__(self):
        return "<SvcOpenBSD name:{} running:{} enabled:{}>".format(
            self.name,
            self.attributes['running'],
            self.attributes['enabled'],
        )

    def fix(self, status):
        if 'enabled' in status.keys_to_fix:
            if self.attributes['enabled'] is False:
                svc_disable(self.node, self.name)
            else:
                svc_enable(self.node, self.name)

        if self.attributes['running'] is False:
            svc_stop(self.node, self.name)
        else:
            svc_start(self.node, self.name)

    def get_canned_actions(self):
        return {
            'stop': {
                'command': "/etc/rc.d/{0} stop".format(self.name),
                'needed_by': {self.id},
            },
            'stopstart': {
                'command': "/etc/rc.d/{0} stop && /etc/rc.d/{0} start".format(self.name),
                'needs': {self.id},
            },
            'restart': {
                'command': "/etc/rc.d/{} restart".format(self.name),
                'needs': {
                    # make sure we don't restart and stopstart simultaneously
                    f"{self.id}:stopstart",
                    # with only the dep on stopstart, we might still end
                    # up reloading if the service itself is skipped
                    # because the stopstart action has cascade_skip False
                    self.id,
                },
            },
        }

    def sdict(self):
        return {
            'enabled': svc_enabled(self.node, self.name),
            'running': svc_running(self.node, self.name),
        }

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if not isinstance(attributes.get('running', True), bool):
            raise BundleError(_(
                "expected boolean for 'running' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))



from datetime import datetime, timedelta
from io import BytesIO
from os import environ
from os.path import normpath
from random import choice
import re
from string import digits, ascii_letters

from . import Fault, STDERR_WRITER


ANSI_ESCAPE = re.compile(r'\x1b[^m]*m')
HIDE_CURSOR = "\033[?25l"
SHOW_CURSOR = "\033[?25h"
VALID_NAME_CHARS = digits + ascii_letters + "-_.+"


def ansi_clean(input_string):
    return ANSI_ESCAPE.sub("", force_text(input_string))


def ansi_wrapper(colorizer):
    if environ.get("BW_COLORS", "1") != "0":
        return colorizer
    else:
        return lambda s, **kwargs: s


@ansi_wrapper
def blue(text):
    return "\033[34m{}\033[0m".format(text)


@ansi_wrapper
def bold(text):
    return "\033[1m{}\033[0m".format(text)


@ansi_wrapper
def cyan(text):
    return "\033[36m{}\033[0m".format(text)


@ansi_wrapper
def italic(text):
    return "\033[3m{}\033[0m".format(text)


@ansi_wrapper
def green(text):
    return "\033[32m{}\033[0m".format(text)


@ansi_wrapper
def red(text):
    return "\033[31m{}\033[0m".format(text)


@ansi_wrapper
def yellow(text):
    return "\033[33m{}\033[0m".format(text)


def cyan_unless_zero(number):
    if number == 0:
        return "0"
    else:
        return cyan(str(number))


def green_unless_zero(number):
    if number == 0:
        return "0"
    else:
        return green(str(number))


def red_unless_zero(number):
    if number == 0:
        return "0"
    else:
        return red(str(number))


def yellow_unless_zero(number):
    if number == 0:
        return "0"
    else:
        return yellow(str(number))


def error_summary(errors):
    if not errors:
        return

    if len(errors) == 1:
        STDERR_WRITER.write(_("\n{x} There was an error, repeated below.\n\n").format(
            x=red("!!!"),
        ))
        STDERR_WRITER.flush()
    else:
        STDERR_WRITER.write(_("\n{x} There were {count} errors, repeated below.\n\n").format(
            count=len(errors),
            x=red("!!!"),
        ))
        STDERR_WRITER.flush()

    for e in errors:
        STDERR_WRITER.write(e)
        STDERR_WRITER.write("\n")
        STDERR_WRITER.flush()


def force_text(data):
    """
    Try to return a text aka unicode object from the given data.
    """
    if isinstance(data, bytes):
        return data.decode('utf-8', 'replace')
    elif isinstance(data, Fault):
        return data.value
    return data


def is_subdirectory(parent, child):
    """
    Returns True if the given child is a subdirectory of the parent.
    """
    parent = normpath(parent)
    child = normpath(child)

    if not parent.startswith("/") or not child.startswith("/"):
        raise ValueError(_("directory paths must be absolute"))

    if parent == child:
        return False

    if parent == "/":
        return True

    return child.startswith(parent + "/")


def mark_for_translation(s):
    return s
_ = mark_for_translation


def randstr(length=24):
    """
    Returns a random alphanumeric string of the given length.
    """
    return ''.join(choice(ascii_letters + digits) for c in range(length))


def validate_name(name):
    """
    Checks whether the given string is a valid name for a node, group,
    or bundle.
    """
    try:
        for char in name:
            assert char in VALID_NAME_CHARS
        assert not name.startswith(".")
    except AssertionError:
        return False
    return True


def wrap_question(title, body, question, prefix=""):
    output = ("{0}\n"
              "{0} ╭─ {1}\n"
              "{0} │\n".format(prefix, title))
    for line in body.splitlines():
        output += "{0} │  {1}\n".format(prefix, line)
    output += ("{0} │\n"
               "{0} ╰─ ".format(prefix) + question)
    return output


class LineBuffer:
    def __init__(self, target):
        self.buffer = b""
        self.record = BytesIO()
        self.target = target if target else lambda s: None

    def close(self):
        self.flush()
        if self.buffer:
            self.record.write(self.buffer)
            self.target(self.buffer)

    def flush(self):
        while b"\n" in self.buffer:
            chunk, self.buffer = self.buffer.split(b"\n", 1)
            self.record.write(chunk + b"\n")
            self.target(chunk + b"\n")

    def write(self, msg):
        self.buffer += msg
        self.flush()


def format_duration(duration, msec=False):
    """
    Takes a timedelta and returns something like "1d 5h 4m 3s".
    """
    components = []
    if duration.days > 0:
        components.append(_("{}d").format(duration.days))
    seconds = duration.seconds
    if seconds >= 3600:
        hours = int(seconds / 3600)
        seconds -= hours * 3600
        components.append(_("{}h").format(hours))
    if seconds >= 60:
        minutes = int(seconds / 60)
        seconds -= minutes * 60
        components.append(_("{}m").format(minutes))
    if seconds > 0 or not components:
        if msec:
            seconds += duration.microseconds / 1000000.0
            components.append(_("{:.3f}s").format(seconds))
        else:
            components.append(_("{}s").format(seconds))
    return " ".join(components)


def format_timestamp(timestamp):
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")


def parse_duration(duration):
    """
    Parses a string like "1d 5h 4m 3s" into a timedelta.
    """
    days = 0
    seconds = 0
    for component in duration.strip().split(" "):
        component = component.strip()
        if component[-1] == "d":
            days += int(component[:-1])
        elif component[-1] == "h":
            seconds += int(component[:-1]) * 3600
        elif component[-1] == "m":
            seconds += int(component[:-1]) * 60
        elif component[-1] == "s":
            seconds += int(component[:-1])
        else:
            raise ValueError(_("{} is not a valid duration string").format(repr(duration)))
    return timedelta(days=days, seconds=seconds)


def toml_clean(s):
    """
    Removes duplicate sections from TOML, e.g.:

        [foo]     <--- this line will be removed since it's redundant
        [foo.bar]
        baz = 1
    """
    lines = list(s.splitlines())
    result = []
    previous = ""
    for line in lines.copy():
        if line.startswith("[") and line.endswith("]"):
            if line[1:].startswith(previous + "."):
                result.pop()
            previous = line[1:-1]
        else:
            previous = ""
        result.append(line)
    return "\n".join(result) + "\n"
